#############################################################################
# Author: schmidtp
# Created on August 06, 2021, 05:06 PM
# Copyright (C) European XFEL GmbH Hamburg. All rights reserved.
#############################################################################

from asyncio import gather, wait_for, TimeoutError as AsyncTimeoutError
from collections import defaultdict
from collections.abc import Hashable
from datetime import datetime
from inspect import ismethod
from traceback import format_exc
from urllib.parse import urlparse
import json
import logging
import re

from tornado.httpclient import AsyncHTTPClient, HTTPError
from tornado.platform.asyncio import AsyncIOMainLoop, to_asyncio_future

from karabo.middlelayer import (
    KaraboError, Device, DeviceClientBase, Descriptor, Hash, Configurable,
    Slot, Node, Type, Schema,
    AccessMode, AccessLevel, Assignment, DaqPolicy, State, Unit,
    UInt16, UInt32, Bool, Float, Double, String, VectorString, VectorHash,
    background, call, callNoWait, setNoWait, sleep, instantiate, slot, coslot,
    get_property, getTopology, getConfiguration, getConfigurationFromPast,
    getConfigurationFromName)
from karabo.middlelayer_api.proxy import ProxyFactory

from ._version import version as deviceVersion
from . import scenes


'''
Device states:
 - INIT: When the device is starting up
 - ACTIVE: When the device is ready to manage a calibration pipeline
 - CHANGING: When the device is actively changing the pipeline configuration
 - UNKNOWN: When the pipeline should be restarted after reconfiguration
 - ERROR: Recoverable error, only allows server restart
 - DISABLED: Unrecoverable error, requires device restart
'''


class ClassIdsNode(Configurable):
    correctionClass = String(
        displayedName='Correction class',
        description='Device class to use for corrections.',
        defaultValue='{}Correction',
        accessMode=AccessMode.INITONLY,
        assignment=Assignment.MANDATORY)

    matcherClass = String(
        displayedName='Matcher class',
        description='Device class to use for matching the stream output of a '
                    'module group.',
        defaultValue='ShmemTrainMatcher',
        accessMode=AccessMode.INITONLY,
        assignment=Assignment.MANDATORY)

    assemblerClass = String(
        displayedName='Assembler class',
        description='Device class to use for assembling the matched output of '
                    'a preview layer.',
        defaultValue='DetectorAssembler',
        accessMode=AccessMode.INITONLY,
        assignment=Assignment.MANDATORY)


class DeviceIdsNode(Configurable):
    correctionSuffix = String(
        displayedName='Correction suffix',
        description='Suffix for correction device IDs. The formatting '
                    'placeholders \'virtualName\', \'index\' and \'group\' '
                    'may be used.',
        defaultValue='CORRECT{index:02d}_{virtualName}',
        accessMode=AccessMode.INITONLY,
        assignment=Assignment.MANDATORY)

    matcherSuffix = String(
        displayedName='Matcher suffix',
        description='Suffix for matching device IDs. The formatting '
                    'placeholder \'group\' may be used.',
        defaultValue='MATCH_G{group}',
        accessMode=AccessMode.INITONLY,
        assignment=Assignment.MANDATORY)

    assemblerSuffix = String(
        displayedName='Assembler suffix',
        description='Suffix for assembler device IDs. The formatting '
                    'placeholder \'layer\' may be used.',
        defaultValue='ASSEMBLE_{layer}',
        accessMode=AccessMode.INITONLY,
        assignment=Assignment.MANDATORY)


class ModuleRow(Configurable):
    virtualName = String(
        displayedName='Virtual name',
        defaultValue='',
        accessMode=AccessMode.RECONFIGURABLE)

    group = UInt32(
        displayedName='Group',
        defaultValue=0,
        accessMode=AccessMode.RECONFIGURABLE)

    aggregator = String(
        displayedName='Aggregator name',
        defaultValue='',
        accessMode=AccessMode.RECONFIGURABLE)

    inputChannel = String(
        displayedName='Input channel',
        defaultValue='',
        accessMode=AccessMode.RECONFIGURABLE)

    inputSource = String(
        displayedName='Input source',
        defaultValue='',
        accessMode=AccessMode.RECONFIGURABLE)


class ModuleGroupRow(Configurable):
    group = UInt32(
        displayedName='Group',
        defaultValue=0,
        accessMode=AccessMode.RECONFIGURABLE)

    deviceServer = String(
        displayedName='Device server',
        defaultValue='',
        accessMode=AccessMode.RECONFIGURABLE)

    withBridge = Bool(
        displayedName='Include bridge?',
        defaultValue=False,
        accessMode=AccessMode.RECONFIGURABLE)

    bridgePort = UInt16(
        displayedName='Bridge port',
        defaultValue=47000,
        minInc=1024,
        maxInc=65353,
        accessMode=AccessMode.RECONFIGURABLE)

    bridgePattern = String(
        displayedName='Bridge pattern',
        options=['PUSH', 'REP', 'PUB'],
        defaultValue='PUSH',
        accessMode=AccessMode.RECONFIGURABLE)


class PreviewLayerRow(Configurable):
    layer = String(
        displayedName='Preview layer',
        defaultValue='',
        accessMode=AccessMode.RECONFIGURABLE)

    outputPipeline = String(
        displayedName='Output pipeline',
        defaultValue='',
        accessMode=AccessMode.RECONFIGURABLE)

    deviceServer = String(
        displayedName='Device server',
        defaultValue='',
        accessMode=AccessMode.RECONFIGURABLE)


class DeviceServerRow(Configurable):
    deviceServer = String(
        displayedName='Device server',
        defaultValue='',
        accessMode=AccessMode.RECONFIGURABLE)

    webserverHost = String(
        displayedName='Webserver host',
        defaultValue='http://',
        accessMode=AccessMode.RECONFIGURABLE)


class WebserverApiNode(Configurable):
    statePollInterval = Double(
        displayedName='Status poll interval',
        description='Time between subsequent polls of the webserver API when '
                    'a device server is expected to reach a particular state.',
        defaultValue=0.5,
        minInc=0.2,
        unitSymbol=Unit.SECOND,
        accessMode=AccessMode.RECONFIGURABLE)

    downTimeout = Double(
        displayedName='Down timeout',
        description='Time to wait for a device server to stop gracefully '
                    'after a \'down\' command and until the \'kill\' command '
                    'is sent.',
        defaultValue=5.0,
        minInc=1.0,
        unitSymbol=Unit.SECOND,
        accessMode=AccessMode.RECONFIGURABLE)

    killTimeout = Double(
        displayedName='Kill timeout',
        description='Time to wait for a device server to stop after a '
                    '\'kill\' command and until the server is deemed '
                    'unrecoverable.',
        defaultValue=5.0,
        minInc=1.0,
        unitSymbol=Unit.SECOND,
        accessMode=AccessMode.RECONFIGURABLE)

    upTimeout = Double(
        displayedName='Up timeout',
        description='Time to wait for a device server to start successfully '
                    'after being down and until the server i deemed '
                    'broken.',
        defaultValue=5.0,
        minInc=1.0,
        unitSymbol=Unit.SECOND,
        accessMode=AccessMode.RECONFIGURABLE)


class RestoredConfigurationRow(Configurable):
    enabled = Bool(
        displayedName='Enabled',
        defaultValue=False,
        accessMode=AccessMode.RECONFIGURABLE)

    memberPattern = String(
        displayedName='Member pattern',
        defaultValue='',
        accessMode=AccessMode.RECONFIGURABLE)

    keyPattern = String(
        displayedName='Key pattern',
        defaultValue='',
        accessMode=AccessMode.RECONFIGURABLE)


class ManagedKeysNode(Configurable):
    pass


class ManagedKeysCloneFactory(ProxyFactory):
    Proxy = ManagedKeysNode
    SubProxy = Configurable
    ProxyNode = Node
    node_factories = dict(Slot=Slot)


class CalibrationManager(DeviceClientBase, Device):
    __version__ = deviceVersion

    interfaces = VectorString(
        displayedName='Device interfaces',
        description='Interfaces implemented by this device.',
        defaultValue=(['DeviceInstantiator']),
        accessMode=AccessMode.READONLY)

    availableScenes = VectorString(
        displayedName='Available scenes',
        displayType='Scenes',
        requiredAccessLevel=AccessLevel.OBSERVER,
        accessMode=AccessMode.READONLY,
        defaultValue=[
            'overview',
            'managed_keys',
            'correction_performance_overview',
            'correction_constant_overview'],
        daqPolicy=DaqPolicy.OMIT)

    @slot
    def requestScene(self, params):
        name = params.get('name', default='overview')

        payload = Hash('success', True)

        if name == 'overview':
            # Assumes there are correction devices known to manager
            scene_data = scenes.manager_device_overview(
                manager_device_id=self.deviceId,
                manager_device_schema=self.getDeviceSchema(),
                correction_device_ids=self._correction_device_ids,
                correction_device_schema=self._correction_device_schema,
                daq_device_ids=self._daq_device_ids,
                domain_device_ids=self._domain_device_ids,
            )
            payload = Hash('success', True, 'name', name, 'data', scene_data)

        elif name == 'correction_performance_overview':
            scene_data = scenes.correction_device_performance_dashboard(
                self._correction_device_ids,
            )
            payload = Hash('success', True, 'name', name, 'data', scene_data)

        elif name == 'correction_constant_overview':
            scene_data = scenes.correction_constant_dashboard(
                self._correction_device_ids,
                self._correction_device_schema,
            )
            payload = Hash('success', True, 'name', name, 'data', scene_data)

        elif name.startswith('browse_schema'):
            if ':' in name:
                prefix = name[len('browse_schema:'):]
            else:
                prefix = 'managedKeys'
            scene_data = scenes.recursive_subschema_scene(
                self.deviceId,
                self.getDeviceSchema(),
                prefix,
            )
            payload = Hash('success', True, 'name', name, 'data', scene_data)

        else:
            payload = Hash('success', False, 'name', name)

        return Hash('type', 'deviceScene',
                    'origin', self.deviceId,
                    'payload', payload)

    detectorType = String(
        displayedName='Detector type',
        description='Type of the detector to manage.',
        options=['AGIPD', 'LPD', 'DSSC', 'Jungfrau', 'ePix100', 'pnCCD',
                 'FastCCD', 'Gotthard2'],
        accessMode=AccessMode.INITONLY,
        assignment=Assignment.MANDATORY)

    detectorIdentifier = String(
        displayedName='Detector identifier',
        description='Name of this detector in CalCat and device ID domain '
                    'for the corresponding data aggregators, if applicable. ',
        accessMode=AccessMode.INITONLY,
        assignment=Assignment.MANDATORY)

    daqDevicesPattern = String(
        displayedName='DAQ devices pattern',
        description='Regexp pattern for the DAQ data aggregators of this '
                    'detector, the formatting placeholder '
                    '\'detector_identifier\' may be used. Leave empty to '
                    'disable DAQ interaction.',
        defaultValue=r'^{detector_identifier}/DET/\d*CH0$',
        accessMode=AccessMode.INITONLY,
        assignment=Assignment.MANDATORY)

    classIds = Node(
        ClassIdsNode,
        displayedName='Class IDs',
        description='class IDs to instantiate for each role. A formatting '
                    'placeholer may be used to indicate the capitalized '
                    'detector type, e.g. Dssc or Jungfrau')

    deviceIds = Node(
        DeviceIdsNode,
        displayedName='Device IDs',
        description='Templates for the member field of the device IDs in the '
                    'pipeline, which is appended to the same device ID root '
                    '(DOMAIN/TYPE) as the manager instance itself.')

    modules = VectorHash(
        displayedName='Modules',
        description='Individual modules constituting this detector. A module '
                    'is the boundary on which correction constants are '
                    'calculated, and in general equivalent to PDUs. Only the '
                    'virtual name and group columns are obligatory, if left '
                    'blank the aggregator name is set to '
                    'f\'{detector_type.upper()}{index:02d}\', the input '
                    'channel is set to '
                    'f\'{detector_identifier}/DET/{index}CH0:output\' and the '
                    'input source '
                    'f\'{detector_identifier}/DET/{index}CH0:xtdf\'.',
        rows=ModuleRow,
        accessMode=AccessMode.RECONFIGURABLE,
        assignment=Assignment.MANDATORY)

    moduleGroups = VectorHash(
        displayedName='Module groups',
        rows=ModuleGroupRow,
        accessMode=AccessMode.RECONFIGURABLE,
        assignment=Assignment.MANDATORY)

    previewLayers = VectorHash(
        displayedName='Preview layers',
        rows=PreviewLayerRow,
        accessMode=AccessMode.RECONFIGURABLE,
        assignment=Assignment.MANDATORY)

    @VectorHash(
        displayedName='Device servers',
        description='WARNING: It is strongly recommended to perform a full '
                    'pipeline restart after changes to this property.',
        rows=DeviceServerRow,
        accessMode=AccessMode.RECONFIGURABLE,
        assignment=Assignment.MANDATORY)
    async def deviceServers(self, value):
        self.deviceServers = value
        self._servers_changed = True

        # Switch to UNKNOWN state to suggest operator to restart pipeline.
        self.state = State.UNKNOWN

    geometryDevice = String(
        displayedName='Geometry device',
        description='Device ID for a geometry device defining the detector '
                    'layout and module positions.',
        accessMode=AccessMode.INITONLY,
        assignment=Assignment.MANDATORY)

    webserverApi = Node(
        WebserverApiNode,
        displayedName='Webserver API',
        description='Configurations for the webserver API to control device '
                    'servers.')

    managedKeyConfiguration = String(
        displayedName='Managed key configuration',
        description='Name of manager device configuration to apply to '
                    'managed keys on startup, leave empty to disable.',
        defaultValue='',
        accessMode=AccessMode.INITONLY)

    @VectorHash(
        displayedName='Restored configurations',
        rows=RestoredConfigurationRow,
        description='Regexp patterns matched against instantiated devices '
                    'to selectively restore the most previous configuration.',
        defaultValue=[Hash(
            'enabled', False,
            'memberPattern', r'MATCH_G\d',
            'keyPattern', '^(sources|zmqConfig|sortSources|mode')],
        accessMode=AccessMode.RECONFIGURABLE)
    async def restoredConfigurations(self, new_configs):
        self.restoredConfigurations = new_configs
        self._restore_config_patterns = {
            re.compile(member_pattern): re.compile(key_pattern)
            for enabled, member_pattern, key_pattern in new_configs.value
            if enabled}

    doNotCompressEvents = Bool(
        requiredAccessLevel=AccessLevel.GOD,
        accessMode=AccessMode.READONLY,
        defaultValue=False,
        daqPolicy=DaqPolicy.OMIT)

    @Slot(
        displayedName='Restart servers',
        allowedStates=[State.UNKNOWN, State.ACTIVE, State.ERROR])
    async def restartServers(self):
        self.state = State.CHANGING
        background(self._restart_servers())

    @Slot(
        displayedName='Instantiate pipeline',
        allowedStates=[State.UNKNOWN, State.ACTIVE])
    async def startInstantiate(self):
        # Slot name is mandated by DeviceInstantiator interface.
        self.state = State.CHANGING
        background(self._instantiate_pipeline())

    managedDevices = VectorString(
        displayedName='Managed devices',
        description='List of currently managed devices.',
        defaultValue=[],
        accessMode=AccessMode.READONLY)

    @Slot(
        displayedName='Discover managed devices',
        description='',
        allowedStates=[State.ACTIVE])
    async def discoverManagedDevices(self):
        self.state = State.CHANGING
        background(self._discover_managed_devices())

    @Slot(
        displayedName='Apply managed values',
        description='Set all managed keys to the values currently active on '
                    'the manager, replacing any manual change.',
        allowedStates=[State.ACTIVE])
    async def applyManagedValues(self):
        background(self._apply_managed_values(
            daq=True, assemblers=True, corrections=True))

    managedKeys = Node(
        ManagedKeysNode,
        displayedName='Managed keys',
        description='Properties and slots managed on devices in the pipeline.')

    # Hardcoded managed keys
    # List of tuple (remote key, local key, descriptor)
    _managed_daq_keys = [
        (
            'DataDispatcher.trainStride', 'daqTrainStride',
            UInt32(displayedName='DAQ train stride',
                   unitSymbol=Unit.COUNT,
                   defaultValue=5,
                   allowedStates=[State.ACTIVE],
                   minInc=1)
        )
    ]

    _managed_assembler_keys = [
        (
            'maxIdle', 'previewMaxIdle',
            Float(displayedName='Preview max idle',
                  unitSymbol=Unit.SECOND,
                  defaultValue=0.5,
                  allowedStates=[State.ACTIVE],
                  minInc=0.0)
        ),
    ]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Take the manager's device ID root (DOMAIN/TYPE) as the root
        # for all devices it manages.
        self._device_id_root = self.deviceId[:self.deviceId.rfind('/')].value

        # Concretized class IDs and device ID templates.
        class_args = (self.detectorType.value.capitalize(),)
        self._class_ids = {}
        self._device_id_templates = {}

        for role in ['correction', 'matcher', 'assembler']:
            self._class_ids[role] = getattr(
                self.classIds, f'{role}Class').value.format(*class_args)
            self._device_id_templates[role] = f'{self._device_id_root}/' + \
                getattr(self.deviceIds, f'{role}Suffix')

        # Concretized pattern for device IDs of data aggregators.
        if self.daqDevicesPattern:
            # Match function of compiled pattern.
            self._is_daq_device = re.compile(self.daqDevicesPattern.format(
                detector_identifier=self.detectorIdentifier.value)).match
        else:
            # Always False.
            self._is_daq_device = bool  # Always returns False.

        # Set of data aggregators associated with the managed detector.
        self._daq_device_ids = set()

        # Set of devices in the same domain as the manager, i.e. having
        # the same device ID root. Not all of these devices are
        # necessarily managed, there may be other persistent devices
        # next to the manager started from projects, e.g. geometry
        # devices or condition checkers.
        self._domain_device_ids = set()

        # Set of managed devices, i.e. in the same domain as the manager
        # and of a class ID used in the classIds node and running in a
        # server in the deviceServers vector.
        self._managed_device_ids = set()

        # Set of correction devices for the managed detector, i.e. being
        # managed and having the specified class ID.
        self._correction_device_ids = set()

        # Set of assembler devices for the managed detector, i.e. being
        # managed and having the specified class ID.
        self._assembler_device_ids = set()

        # Mapping of device ID pattern to key pattern to selectively
        # restore past configuration for.
        self._restore_config_patterns = dict()

        # Task object to update managed devices list.
        self._managed_devices_updater = None

    async def onInitialization(self):
        self.state = State.INIT
        await self._async_init()

    @coslot
    async def slotInstanceNew(self, instance_id, info):
        # A new instance has appeared in the topology.

        await super().slotInstanceNew(instance_id, info)

        if info['type'] == 'device':
            self._check_new_device(instance_id, info['serverId'],
                                   info['classId'])
            self._update_managed_devices()

    @slot
    def slotInstanceGone(self, instance_id, info):
        # An instance is gone from the topology.

        super().slotInstanceGone(instance_id, info)

        if info['type'] == 'device':
            self._daq_device_ids.discard(instance_id)
            self._domain_device_ids.discard(instance_id)
            self._managed_device_ids.discard(instance_id)
            self._correction_device_ids.discard(instance_id)
            self._assembler_device_ids.discard(instance_id)
            self._update_managed_devices()

    async def _async_init(self):
        # Set-up Tornado.
        if hasattr(AsyncIOMainLoop, 'initialized'):
            if not AsyncIOMainLoop.initialized():
                AsyncIOMainLoop().install()

        self._http_client = AsyncHTTPClient()

        # Check device servers and initialize webserver access.
        await self._check_servers()

        # Inject schema for configuration of managed devices.
        await self._inject_managed_keys()

        # Populate the device ID sets with what's out there right now.
        await self._check_topology()

        if self.state == State.INIT:
            self._set_status('Calibration manager ready')
            self.state = State.ACTIVE

    def _check_new_device(self, device_id, server_id, class_id):
        if device_id == self.deviceId:
            # Ignore manager itself, it typically sees its own instance
            # going up first.
            return

        elif class_id == 'DataAggregator' and self._is_daq_device(device_id):
            # This device is a data aggregator belonging to the detector
            # installation
            self._daq_device_ids.add(device_id)

        elif device_id.startswith(self._device_id_root):
            # In our domain.
            self._domain_device_ids.add(device_id)

            if (
                server_id not in self._server_api_names
                or class_id not in self._class_ids.values()
            ):
                # NOT a managed device.
                return

            # A managed device.
            self._managed_device_ids.add(device_id)

            if class_id == self._class_ids['correction']:
                # This device is also a correction device.
                self._correction_device_ids.add(device_id)

            elif class_id == self._class_ids['assembler']:
                # This device is also an assembler
                self._assembler_device_ids.add(device_id)

    async def _check_topology(self):
        for i in range(10):
            try:
                devices = getTopology()['device']
            except AttributeError:
                # Not ready yet
                await sleep(1.0)
            else:
                if not devices:
                    await sleep(1.0)
                else:
                    break
        else:
            self._set_fatal(f'Topology not available after {i+1} tries')
            return

        self.logger.debug(f'Topology arrived after {i+1} tries')

        for device_id in devices:
            self._check_new_device(device_id, devices[device_id, 'serverId'],
                                   devices[device_id, 'classId'])

        self._update_managed_devices(True)

    async def _discover_managed_devices(self):
        self._daq_device_ids.clear()
        self._domain_device_ids.clear()
        self._managed_device_ids.clear()
        self._correction_device_ids.clear()
        self._assembler_device_ids.clear()

        await self._check_topology()

        self.state = State.ACTIVE

    async def _delayed_managed_devices_update(self):
        await sleep(1.0)  # Throttle updates to at most once a second.
        self.managedDevices = sorted(self._managed_device_ids)
        self._managed_devices_updater = None  # Clear task again.

    def _update_managed_devices(self, forced=False):
        if self._managed_devices_updater is not None:
            # Update already in progress, ignore.
            return

        if forced or len(self._managed_device_ids) != len(self.managedDevices):
            # Trigger an update either if forced or the number of
            # devices changed.
            self._managed_devices_updater = background(
                self._delayed_managed_devices_update())

    async def _get_shared_keys(self, device_ids, keys):
        """Find the most common property values on devices."""

        key_values = defaultdict(list)

        for device_id in device_ids:
            try:
                config = await wait_for(getConfiguration(device_id),
                                        timeout=2.0)
            except AsyncTimeoutError:
                # Ignore this device if the configuration can no longer
                # be obtained.
                continue

            for key in keys:
                value = config[key]

                if isinstance(value, Hashable):
                    # Value must be hashable to determine the most
                    # common one below.
                    key_values[key].append(value)

        return {key: max(set(values), key=values.count) for key, values
                in key_values.items()}

    async def _add_managed_keys(self, managed_hash, managed_keys, owner,
                                device_ids):
        """Add hardcoded managed keys."""

        remote_keys = []

        for remote_key, local_key, descr in managed_keys:
            remote_keys.append(remote_key)
            _, attrs = descr.toSchemaAndAttrs(None, None)

            managed_hash[local_key] = 0
            managed_hash[local_key, ...].update(
                attrs, __owner=owner, __remote=remote_key)

        return await self._get_shared_keys(device_ids, remote_keys) \
            if device_ids else {}

    async def _inject_managed_keys(self):
        """Attempt to retrieve the correction device's schema and insert
        part of it as managed keys, plus some hardcoded keys for other
        devices types.
        """

        correction_device_servers = [
            server for _, server, _, _, _ in self.moduleGroups.value]

        up_corr_servers = await self._get_servers_in_state(
            'up', servers=correction_device_servers)

        if up_corr_servers:
            # At least one correction server is up.
            corr_server = next(iter(up_corr_servers))
        else:
            # None of the correction servers is up, try to start the
            # first one.
            corr_server = correction_device_servers[0]

            try:
                await wait_for(self._ensure_server_state(corr_server, 'up'),
                               timeout=self.webserverApi.upTimeout.value)
            except AsyncTimeoutError:
                self._set_fatal(f'Could not bring up correction device server '
                                f'`{corr_server}` for schema retrieval')
                return

        # Obtain the device schema from a correction device server.
        managed_schema, _, _ = await call(corr_server, 'slotGetClassSchema',
                                          self._class_ids['correction'])

        # Save this for scene generation later.
        self._correction_device_schema = Schema()
        self._correction_device_schema.copy(managed_schema)

        if managed_schema.name != self._class_ids['correction']:
            self._set_fatal(
                f'Correction class ID `{self._class_ids["correction"]}` not '
                f'known or loadable by device server `{corr_server}`')
            return

        managed_hash = managed_schema.hash
        # full paths and intermediate node paths
        managed_paths = set(managed_schema.filterByTags('managed').paths())
        # only leaf nodes (actual things to set)
        managed_keys = set(path for path in managed_paths
                           if not isinstance(managed_hash[path], Hash))

        # Reduce the correction schema to the managed paths.
        for path in managed_hash.paths():
            if path in managed_paths:
                # Set owner and remote name (identical for corrections).
                managed_hash[path, ...].update(
                    __owner='corrections', __remote=path)
            else:
                # Remove unmanaged path.
                del managed_hash[path]

        # Retrieve any previous values already on running devices in
        # order to update the defaultValue attribute in the schema just
        # before injection.
        prev_values = await self._get_shared_keys(
            self._correction_device_ids, managed_keys)

        # Add in managed DAQ keys, if enabled.
        if self.daqDevicesPattern:
            prev_values.update(await self._add_managed_keys(
                managed_hash, self._managed_daq_keys, 'daq',
                self._daq_device_ids))

        # Add in managed assembler keys.
        prev_values.update(await self._add_managed_keys(
            managed_hash, self._managed_assembler_keys, 'assemblers',
            self._assembler_device_ids))

        if self.managedKeyConfiguration:
            config_name = self.managedKeyConfiguration.value

            # Try to obtain the specified manager configuration to apply
            # its values of managed keys on top of the current values.
            try:
                named_config = (await getConfigurationFromName(
                    self.deviceId, config_name))['managedKeys']
            except KaraboError as e:
                self.logger.warn(f'Failed receiving named manager '
                                 f'configuration `{config_name}`: {e}')
            except KeyError:
                self.logger.warn(f'Missing `managedKeys` on named manager '
                                 f'configuration `{config_name}`')
            else:
                # Pick out the keys managed now and check for hashable
                # values (e.g. excluding slots).
                prev_values.update({
                    key: named_config[key]
                    for key in set(named_config.paths()) & managed_keys
                    if isinstance(named_config[key], Hashable)})

        # Retrieve the attributes on the current managed node. The
        # original implementation of toSchemaAndAttrs in the Node's
        # superclass Descriptor is used to avoid Node-specific
        # attributes in the attributes that are not valid in the
        # property definition.
        # The value are then obtained from the Node object again since
        # enums are converted to their values by toSchemaAndAttrs, which
        # in turn is not valid for property definition.
        _, attrs = Descriptor.toSchemaAndAttrs(self.__class__.managedKeys,
                                               None, None)
        managed_node_attrs = {key: getattr(self.__class__.managedKeys, key)
                              for key in attrs.keys()}

        # Build a proxy from the managed schema, and create a new node
        # based on it with the original node attributes. This code is
        # heavily inspired by deviceClone.
        managed_node = Node(
            ManagedKeysCloneFactory.createProxy(managed_schema),
            **managed_node_attrs)

        # Walk the managed tree to and sanitize all descriptors to our
        # specifications.
        def _sanitize_node(parent, tree, prefix=''):
            for leaf_key, value in tree.items():
                # Fetch the descriptor class, not its instance!
                descr = getattr(parent.cls, leaf_key)

                local_key = f'{prefix}.{leaf_key}' if prefix else leaf_key
                remote_key = managed_hash[local_key, '__remote']

                if isinstance(descr, Node):
                    _sanitize_node(descr, value, local_key)

                elif isinstance(descr, Slot):
                    async def _managed_slot_called(parent,
                                                   remote_key=remote_key):
                        background(self._call_on_corrections(remote_key))

                    _managed_slot_called.__name__ = f'managedKeys.{local_key}'
                    descr.__call__(_managed_slot_called)

                    # Managed slots can only be called in the ACTIVE
                    # state.
                    descr.allowedStates = [State.ACTIVE]

                elif isinstance(descr, Type):
                    # Regular property.

                    if descr.accessMode == AccessMode.RECONFIGURABLE:
                        # Add a callback only if the original descriptor
                        # is reconfigurable.

                        async def _managed_prop_changed(
                            parent, new_value,
                            leaf_key=leaf_key,
                            remote_key=remote_key,
                            setter=getattr(self, '_set_on_' + managed_hash[
                                local_key, '__owner'])
                        ):
                            setattr(parent, leaf_key, new_value)

                            if self.state != State.INIT:
                                # Do not propagate updates during injection.
                                background(setter(remote_key, new_value))

                        descr.__call__(_managed_prop_changed)

                    # Managed properties are always optional,
                    # reconfigurable and may only be changed in the
                    # ACTIVE state.
                    descr.assignment = Assignment.OPTIONAL
                    descr.accessMode = AccessMode.RECONFIGURABLE
                    descr.allowedStates = [State.ACTIVE]

                    try:
                        # If there's been a previous value before
                        # injection, use it.
                        descr.defaultValue = prev_values[remote_key]
                    except KeyError:
                        pass
                else:
                    self.logger.warn(f'Encountered unknown descriptor type '
                                     f'{type(descr)}')

        _sanitize_node(managed_node, managed_hash)

        # Inject the newly prepared node for managed keys.
        self.__class__.managedKeys = managed_node
        await self.publishInjectedParameters()
        self._managed_correction_keys = managed_keys

        self.logger.debug('Managed schema injected')

    def _set_status(self, text, level=logging.INFO):
        """Add and log a status message.

        Suppresses throttling from the gui server.
        """

        self.status = text
        self.doNotCompressEvents = not self.doNotCompressEvents
        self.logger.log(level, text)

    def _set_error(self, text):
        """Set the device into error state and log an error message."""

        self.state = State.ERROR
        self._set_status(text, level=logging.ERROR)

    def _set_fatal(self, text):
        """Set the device into disabled state and log an error message."""

        self.state = State.DISABLED
        self._set_status(text, level=logging.CRITICAL)

    def _set_exception(self, text, e):
        """Set the device into error upon an exception."""

        return self._set_error(f'{text}\n{format_exc()[:-1]}')

    def _get_server_api_url(self, name):
        """Get API URL for a device server."""

        return '{host}/api/servers/{api_name}.json'.format(
            host=self._server_hosts[name],
            api_name=self._server_api_names[name])

    async def _get_server_info(self, name):
        """Get info for a device server.

        Args:
            name (str): Device server (Karabo) name.

        Returns:
            (dict) Information returned by webserver for this server.
        """

        try:
            reply = await to_asyncio_future(self._http_client.fetch(
                self._get_server_api_url(name), method='GET'))
        except (ConnectionError, HTTPError) as e:
            raise RuntimeError(
                f'Failed retrieving server info of {name} on '
                f'{urlparse(self._server_hosts[name]).hostname}: {e}')

        body = json.loads(reply.body)

        if not body['success']:
            raise RuntimeError(f'Request to retrieve server info for {name} '
                               f'failed')

        self.logger.debug(f'Obtained server info for {name}: '
                          f'{body["servers"][0]}')

        return body['servers'][0]

    async def _get_servers_in_state(self, state, servers=None):
        """List all servers in a particular state.

        Args:
            state (str): State to filter for, e.g. 'up' or 'down'.

        Returns:
            (list[str]) List of server names found in the passed state.
        """

        servers = await gather(*[self._get_server_info(name)
                                 for name in self._server_hosts.keys()
                                 if servers is None or name in servers])

        return {server['karabo_name'] for server in servers
                if server['status'].startswith(state)}

    async def _ensure_server_state(self, name, state, command=None):
        """Sets and waits for a a server state to be reached.

        Args:
            name (str): Device server (Karabo) name.
            state (str): State to set and wait for, e.g. 'up' or 'down'.
            command (str, optional): Command to use for reaching the
                state, same as state of omitted.
        """

        if command is None:
            command = state

        try:
            reply = await to_asyncio_future(self._http_client.fetch(
                self._get_server_api_url(name), method='PUT',
                body=json.dumps({'server': {'command': command}})))
        except (ConnectionError, HTTPError) as e:
            raise RuntimeError(
                f'Failed sending `{command}` to {name} on '
                f'{urlparse(self._server_hosts[name]).hostname}: {e}')

        if not json.loads(reply.body)['success']:
            raise RuntimeError('Command `{command}` for {name} failed')

        while True:
            await sleep(self.webserverApi.statePollInterval.value)
            info = await self._get_server_info(name)

            if info['status'].startswith(state):
                return

    async def _check_servers(self):
        """Validate device server configuration."""

        if not self._servers_changed:
            return True

        # Mapping of servers to device servers.
        self._server_hosts = {server: host for server, host
                              in self.deviceServers.value}

        # Mapping of "Karabo names" (with capitalization and slashes)
        # to "API names" (all lower case and only underscores).
        self._server_api_names = {}

        # Build a mapping of hosts to server names.
        hosts = defaultdict(set)
        for name, host in self._server_hosts.items():
            hosts[host].add(name)

        # List of errors found during device server check.
        err = []

        # Query all servers on each host and check whether the ones we
        # need are in there, and obtain their API names.
        for host, req_names in hosts.items():
            # Retrieve hostname for nice error messages.
            hostname = urlparse(self._server_hosts[name]).hostname

            try:
                reply = await to_asyncio_future(self._http_client.fetch(
                    f'{host}/api/servers.json', method='GET'))
            except (ConnectionError, HTTPError) as e:
                err.append(f'- {e.__class__.__name__} when retrieving server '
                           f'list on {hostname}: {e}')
                continue

            body = json.loads(reply.body)

            if not body['success']:
                err.append(f'- Request to retrieve server list failed on '
                           f'{hostname}')
                continue

            servers = {s['karabo_name']: s for s in body['servers']
                       if s['karabo_name'] in req_names}

            if len(servers) != len(req_names):
                err.append(f'- Device servers missing on {hostname}: '
                           + ', '.join(req_names - servers.keys()))

            servers_in_error = {s['karabo_name'] for s in servers.values()
                                if 'error' in s['status']}

            if servers_in_error:
                err.append(f'- Device servers on {hostname} are in error state'
                           + ', '.join(servers_in_error))
                continue

            servers_disabled = {s['karabo_name'] for s in servers.values()
                                if not s['control_allowed']}

            if servers_disabled:
                err.append(f'- Device servers on {hostname} not controllable '
                           f'via webserver: ' + ', '.join(servers_disabled))
                continue

            for name, info in servers.items():
                self._server_api_names[name] = info['name']

        if err:
            self._set_error('One or more device server problems were '
                            'detected:\n' + '\n'.join(err))
            return False
        else:
            return True

    async def _restart_servers(self):
        """Restart all managed device servers."""

        if not await self._check_servers():
            return

        try:
            # Find all servers in need of being brought down.
            up_names = await self._get_servers_in_state('up')

            # Bring down all servers which are up.
            if up_names:
                try:
                    await wait_for(gather(
                        *[self._ensure_server_state(name, 'down')
                          for name in up_names]),
                        timeout=self.webserverApi.downTimeout.value)
                except AsyncTimeoutError:
                    # Narrow down to the list of servers STILL up.
                    up_names = await self._get_servers_in_state('up')
                else:
                    up_names.clear()

            # If some servers are still up, go for the kill.
            if up_names:
                self.logger.warn('Some servers still up after waiting for '
                                 f'downTimeout: {up_names}')

                try:
                    await wait_for(gather(
                        *[self._ensure_server_state(name, 'down', 'kill')
                          for name in up_names]),
                        timeout=self.webserverApi.killTimeout.value)
                except AsyncTimeoutError:
                    # All hope is lost.
                    up_names = await self._get_servers_in_state('up')
                    return self._set_error(
                        'One or more device servers could not be brought '
                        'down, even via kill:' + ', '.join(up_names))

            # Bring all servers up again.
            try:
                await wait_for(gather(
                    *[self._ensure_server_state(name, 'up')
                      for name in self._server_hosts.keys()]),
                    timeout=self.webserverApi.upTimeout.value)
            except AsyncTimeoutError:
                down_names = await self._get_servers_in_state('down')
                return self._set_error('One or more device servers could not '
                                       'brought up: ' + ', '.join(down_names))

        except RuntimeError as e:
            return self._set_exception(
                'Request unexpectedly failed during restart procedure', e)

        # Wait a moment for good measure until the servers are
        # actually up. If we try to instantiate to quickly after the
        # servers reported running, instantiate() blocks forever.
        await sleep(5.0)

        if self._domain_device_ids:
            # If there are still devices in this set, log a warning.
            self.logger.warn('Some devices left in manager domain after '
                             'device servers have been restarted: '
                             + ', '.join(self._domain_device_ids))

        self._set_status('All device servers restarted')
        self.state = State.ACTIVE

    async def _instantiate_device(self, server, class_id, device_id, config):
        """Instantiate a single device.

        Small wrapper around karabo.middlelayer.instantiate for error
        handling and logging.
        """

        if device_id in self._domain_device_ids:
            self.logger.debug(f'Skipped instantiation of already existing '
                              f'device {device_id}')
            return True

        # Find any applicable key patterns to restore configuration for.
        member_str = device_id[device_id.rfind('/') + 1:]
        key_patterns = [key_pattern for member_pattern, key_pattern
                        in self._restore_config_patterns.items()
                        if member_pattern.match(member_str)]

        if key_patterns:
            try:
                # Try to obtain most recent configuration.
                old_config = await wait_for(getConfigurationFromPast(
                    device_id, datetime.now().isoformat()), 15.0)
            except AsyncTimeoutError:
                self.logger.warn(f'Timeout receiving previous configuration '
                                 f'for {device_id}')
            except KaraboError as e:
                self.logger.warn(f'Failed receiving previous configuration '
                                 f'for {device_id}: {e}')
            else:
                # Match all keys against the found patterns.
                keys = {key for key in old_config.paths()
                        if any((key_pattern.match(key)
                                for key_pattern in key_patterns))}
                self.logger.debug(f'Keys restored on {device_id}: '
                                  + ', '.join(keys))

                config.update({key: old_config[key] for key in keys})

        try:
            msg = await wait_for(instantiate(
                server, class_id, device_id, config), 20.0)
        except AsyncTimeoutError:
            self._set_error(f'Instantiation timeout on {device_id}')
            return False
        except KaraboError as e:
            self._set_exception(f'Instantiation error on {device_id}', e)
            return False

        self.logger.debug(f'Instantation result for {device_id}: {msg}')

        return True

    async def _instantiate_pipeline(self):
        """Instantiate all managed devices."""

        if not await self._check_servers():
            return

        # Make sure all servers are up.
        try:
            up_servers = await self._get_servers_in_state('up')

            if up_servers != self._server_api_names.keys():
                self.state = State.ACTIVE
                return self._set_status('One or more device servers are not '
                                        'up, restart servers first.')
        except RuntimeError as e:
            return self._set_error('Request unexpectedly failed while '
                                   'checking device server state', e)

        # Servers by group and layer.
        server_by_group = {group: server for group, server, _, _, _,
                           in self.moduleGroups.value}
        server_by_layer = {layer: server for layer, _, server
                           in self.previewLayers.value}

        all_req_servers = set(server_by_group.values()).union(
            server_by_layer.values())

        if all_req_servers != up_servers:
            return self._set_error('One or more device servers are not '
                                   'listed in the device servers '
                                   'configuration')

        # Instantiate modules.
        modules_by_group = defaultdict(list)
        correct_device_id_by_module = {}
        input_source_by_module = {}

        awaitables = []
        for index, row in enumerate(self.modules.value):
            vname, group, aggregator, input_channel, input_source = row

            modules_by_group[group].append(vname)
            device_id = self._device_id_templates['correction'].format(
                virtualName=vname, index=index, group=group)
            correct_device_id_by_module[vname] = device_id

            if not aggregator:
                aggregator = f'{self.detectorType.upper()}{index:02d}'

            detector_id = self.detectorIdentifier.value
            daq_device_id = f'{detector_id}/DET/{index}CH0'

            if not input_channel:
                input_channel = f'{daq_device_id}:output'

            if not input_source:
                input_source = f'{daq_device_id}:xtdf'

            input_source_by_module[vname] = input_source

            config = Hash(
                'constantParameters.detectorName', detector_id,
                'constantParameters.karaboDa', aggregator,
                'dataInput.connectedOutputChannels', [input_channel],
                'fastSources', [input_source],
                'geometryDevice', self.geometryDevice.value
            )

            # Add managed keys.
            for key in self._managed_correction_keys:
                value = get_property(self, f'managedKeys.{key}')

                if not ismethod(value):
                    config[key] = value.value

            awaitables.append(self._instantiate_device(
                server_by_group[group],
                self._class_ids['correction'],
                device_id,
                config
            ))

        # Instantiate group matchers with optional bridge enabled.
        for row in self.moduleGroups.value:
            group, server, with_bridge, bridge_port, bridge_pattern = row

            sources = [Hash('select', True,
                            'source', '{}@{}:dataOutput'.format(
                                input_source_by_module[vname],
                                correct_device_id_by_module[vname])
                            )
                       for vname in modules_by_group[group]]
            config = Hash('sources', sources)

            if with_bridge:
                config['zmqConfig'] = [
                    Hash('pattern', bridge_pattern, 'hwm', 1,
                         'port', bridge_port)
                ]

            awaitables.append(self._instantiate_device(
                server,
                self._class_ids['matcher'],
                self._device_id_templates['matcher'].format(group=group),
                config
            ))

        # Instantiate preview layer assemblers.
        for layer, output_pipeline, server in self.previewLayers.value:
            sources = [Hash('select', True,
                            'source', f'{device_id}:{output_pipeline}')
                       for (_, device_id)
                       in correct_device_id_by_module.items()]
            config = Hash('sources', sources,
                          'geometryDevice', self.geometryDevice.value)

            for remote_key, local_key, _ in self._managed_assembler_keys:
                value = get_property(self, f'managedKeys.{local_key}')

                if not ismethod(value):
                    config[remote_key] = value.value

            awaitables.append(self._instantiate_device(
                server,
                self._class_ids['assembler'],
                self._device_id_templates['assembler'].format(layer=layer),
                config
            ))

        # Perform instantation.
        await gather(*awaitables)

        # Force managed DAQ settings.
        await self._apply_managed_values(daq=True)

        # Ask the geometry device to re-send its geometry.
        callNoWait(self.geometryDevice.value, 'sendGeometry')

        self._set_status('All devices instantiated')
        self.state = State.ACTIVE

    async def _apply_managed_values(self, daq=False, assemblers=False,
                                    corrections=False):
        """Apply all managed keys to local values."""

        if daq and self.daqDevicesPattern:
            # Only apply if DAQ management is enabled.
            for remote_key, local_key, _ in self._managed_daq_keys:
                await self._set_on_daq(
                    remote_key, get_property(self, f'managedKeys.{local_key}'))

        if assemblers:
            for remote_key, local_key, _ in self._managed_assembler_keys:
                await self._set_on_assembler(
                    remote_key, get_property(self, f'managedKeys.{local_key}'))

        if corrections:
            for key in self._managed_correction_keys:
                value = get_property(self, f'managedKeys.{key}')
                if not ismethod(value):
                    await self._set_on_corrections(key, value)

    def _call(self, device_ids, slot, *args):
        """Call the same slot on a list of devices."""

        for device_id in device_ids:
            callNoWait(device_id, slot, *args)
            self.logger.debug(f'Called {device_id}.{slot}{args}')

    async def _call_on_corrections(self, slot, *args):
        """Call a slot on all correction devices."""

        if self._correction_device_ids:
            self._call(self._correction_device_ids, slot, *args)
            self.logger.info(f'Called <CORR>.{slot}{args}')

    def _set(self, device_ids, key, value):
        """Set the same property on a list of devices."""

        for device_id in device_ids:
            setNoWait(device_id, key, value.value)
            self.logger.debug(f'Set {device_id}.{key} to {value.value}')

    async def _set_on_daq(self, key, value):
        """Set a property on all DAQ devices."""

        if self._daq_device_ids:
            self._set(self._daq_device_ids, key, value)
            self.logger.info(f'Set <DAQ>.{key} to {value}')

    async def _set_on_corrections(self, key, value):
        """Set a property on all correction devices."""

        if self._correction_device_ids:
            self._set(self._correction_device_ids, key, value)
            self.logger.info(f'Set <CORR>.{key} to {value}')

    async def _set_on_assemblers(self, key, value):
        """Set a property on all preview assemblers."""

        if self._assembler_device_ids:
            self._set(self._assembler_device_ids, key, value)
            self.logger.info(f'Set <ASSEMBLE>.{key} to {value}')
