from importlib.metadata import entry_points

from karabo.bound import (
    KARABO_CLASSINFO,
    NODE_ELEMENT,
    OUTPUT_CHANNEL,
    STRING_ELEMENT,
    VECTOR_BOOL_ELEMENT,
    ChannelMetaData,
    Epochstamp,
    Hash,
    Schema,
    Timestamp,
    Trainstamp,
)

from TrainMatcher import TrainMatcher

from ._version import version as deviceVersion
from . import utils, geom_utils
from .arbiter_kernels.base_kernel import KernelWarning

my_schema = Schema()
(
    NODE_ELEMENT(my_schema)
    .key("data")
    .commit(),

    VECTOR_BOOL_ELEMENT(my_schema)
    .key("data.dataFramePattern")
    .assignmentOptional()
    .defaultValue([])
    .commit(),
)


kernels = [
    kernel_class.load()
    for kernel_class in entry_points().get("calng.arbiter_kernel", [])
]
kernel_choice = {
    kernel_class.__name__: kernel_class for kernel_class in kernels
}


@KARABO_CLASSINFO("FrameSelectionArbiter", deviceVersion)
class FrameSelectionArbiter(TrainMatcher.TrainMatcher):
    @staticmethod
    def expectedParameters(expected):
        (
            STRING_ELEMENT(expected)
            .key("geometryDevice")
            .displayedName("Geometry device")
            .description(
                "The the device which will provide geometry. The device is expected to "
                "have current geometry as a string (for now using custom non-robust "
                "serialization scheme) element named serializedGeometry."
            )
            .assignmentMandatory()
            .commit(),

            NODE_ELEMENT(expected)
            .key("frameSelection")
            .displayedName("Frame selection")
            .commit(),

            STRING_ELEMENT(expected)
            .key("frameSelection.kernelChoice")
            .displayedName("Kernel to use")
            .assignmentMandatory()
            .options(",".join(kernel_choice.keys()))
            .reconfigurable()
            .commit(),

            STRING_ELEMENT(expected)
            .key("frameSelection.kernelState")
            .setSpecialDisplayType("State")
            .displayedName("Kernel state")
            .readOnly()
            .initialValue("OFF")
            .commit(),

            NODE_ELEMENT(expected)
            .key("frameSelection.kernels")
            .displayedName("Kernels")
            .commit(),

            OUTPUT_CHANNEL(expected)
            .key("output")
            .dataSchema(my_schema)
            .commit(),
        )
        for kernel_class in kernel_choice.values():
            kernel_class.extend_device_schema(
                expected, f"frameSelection.kernels.{kernel_class.__name__}"
            )

    def __init__(self, config):
        super().__init__(config)
        self.registerSlot(self.slotReceiveGeometry)

    def initialization(self):
        super().initialization()
        self.warning_context = utils.WarningContextSystem(
            self,
            on_success={"frameSelection.kernelState": "ON"},
        )
        with self.warning_context("frameSelection.kernelState", KernelWarning.INIT):
            self._kernel_class = kernel_choice[self.get("frameSelection.kernelChoice")]
            self._kernel = self._kernel_class(
                self._parameters[
                    f"frameSelection.kernels.{self._kernel_class.__name__}"
                ]
            )
            self._kernel._device = self

        self._geometry = None
        if self.get("geometryDevice"):
            self.signalSlotable.connect(
                self.get("geometryDevice"),
                "signalNewGeometry",
                "",  # slot device ID (default: self)
                "slotReceiveGeometry",
            )
        self.start()  # Auto-start this type of matcher.

    def on_matched_data(self, train_id, sources):
        # TODO: robust frame deduction
        num_frames = None
        for (data, _) in sources.values():
            if not data.has("image.cellId"):
                continue
            num_frames = data.get("image.cellId").size
            break
        else:
            # TODO: also use warning context
            self.log.WARN("Unable to figure out number of frames")

        with self.warning_context(
            "frameSelection.kernelState", KernelWarning.PROCESSING
        ):
            decision = self._kernel.consider(train_id, sources, num_frames)
        if isinstance(decision, Hash):
            assert decision.has("data.dataFramePattern")
            decision["data.dataFramePattern"] = list(
                map(bool, decision["data.dataFramePattern"])
            )
            result = decision
        else:
            result = Hash()
            # TODO: avoid recasting
            result["data.dataFramePattern"] = list(map(bool, decision))
        self.output.write(
            result,
            ChannelMetaData(
                f"{self.getInstanceId()}:output",
                Timestamp(Epochstamp(), Trainstamp(train_id)),
            ),
        )
        self.output.update()
        self.rate_out.update()

    def start(self):
        self.set("frameSelection.kernelState", "OFF")
        self.warning_context.reset()
        super().start()

    def preReconfigure(self, conf):
        super().preReconfigure(conf)
        self._reinstantiate_arbiter_kernel = False
        if conf.has("frameSelection.kernelChoice"):
            self.log.INFO("Switching frame selection kernel")
            self._kernel_class = kernel_choice[conf.get("frameSelection.kernelChoice")]
            self._reinstantiate_arbiter_kernel = True
        elif conf.has(f"frameSelection.kernels.{self._kernel_class.__name__}"):
            self.log.INFO("Reconfiguring frame selection kernel")
            # TODO: update instead of rebuild for complex kernels
            # (decide based on reconfigurability?)
            self._reinstantiate_arbiter_kernel = True

    def postReconfigure(self):
        super().postReconfigure()
        if self._reinstantiate_arbiter_kernel:
            self._kernel = self._kernel_class(
                self._parameters[
                    f"frameSelection.kernels.{self._kernel_class.__name__}"
                ]
            )
            self._reinstantiate_arbiter_kernel = False

    def slotReceiveGeometry(self, device_id, serialized_geometry):
        self.log.INFO(f"Received geometry from {device_id}")
        try:
            self._geometry = geom_utils.deserialize_geometry(serialized_geometry)
        except Exception as e:
            self.log.WARN(f"Failed to deserialize geometry; {e}")
