import numpy as np
from karabo.bound import (
    KARABO_CLASSINFO,
    OUTPUT_CHANNEL,
    OVERWRITE_ELEMENT,
    STRING_ELEMENT,
    ChannelMetaData,
    Dims,
    Encoding,
    Epochstamp,
    Hash,
    ImageData,
    Schema,
    State,
    Timestamp,
    Trainstamp,
)
from TrainMatcher import TrainMatcher

from . import scenes, schemas
from . import shmem_utils, utils
from ._version import version as deviceVersion


@KARABO_CLASSINFO("Gotthard2Assembler", deviceVersion)
class Gotthard2Assembler(TrainMatcher.TrainMatcher):
    @staticmethod
    def expectedParameters(expected):
        (
            OVERWRITE_ELEMENT(expected)
            .key("availableScenes")
            .setNewDefaultValue(["overview", "trainMatcherScene"])
            .commit(),

            OUTPUT_CHANNEL(expected)
            .key("output")
            .dataSchema(schemas.preview_schema(wrap_image_in_imagedata=False))
            .commit(),

            STRING_ELEMENT(expected)
            .key("geometryDevice")
            .displayedName("[Unused] Geometry device")
            .description(
                "Assembly of 25 μm GOTTHARD-II does not require geometry per se. Just "
                "make sure to, in the sources list, list the master module first and "
                "the slave module second."
            )
            .assignmentOptional()
            .defaultValue("")
            .commit(),
        )

    def initialization(self):
        super().initialization()
        sources = self.get("sources")
        if len(sources) != 2:
            self.log.WARN("Must have exactly two GOTTHARD-II sources for 25 μm mode")
            self.updateState(State.ERROR)
        self._primary_source, self._secondary_source = [
            row["source"].partition("@")[0] for row in self.get("sources")
        ]
        self._shmem_handler = shmem_utils.ShmemCircularBufferReceiver()
        self._interleaving_buffer = np.ma.empty(0, dtype=np.float32)
        self._wrap_in_imagedata = False
        self.start()

    def requestScene(self, params):
        scene_name = params.get("name", default="")
        if scene_name == "trainMatcherScene":
            params["name"] = "scene"
            return super().requestScene(params)
        if scene_name != "overview":
            self.log.INFO(
                f"Unexpected scene name {scene_name} requested, "
                "will just return overview"
            )
        payload = Hash("name", scene_name, "success", True)
        payload["data"] = scenes.gotthard2_assembler_overview(
            device_id=self.getInstanceId(),
            schema=self.getFullSchema(),
        )
        self.reply(
            Hash(
                "type",
                "deviceScene",
                "origin",
                self.getInstanceId(),
                "payload",
                payload,
            )
        )

    def on_matched_data(self, train_id, sources):
        for (data, _) in sources.values():
            self._shmem_handler.dereference_shmem_handles(data)

        # data hash format varies between dataOutput and preview.output*
        if sources[self._primary_source][0].has("data.adc"):
            data_1 = sources[self._primary_source][0]["data.adc"]
            data_2 = sources[self._secondary_source][0]["data.adc"]
        else:
            data_1 = sources[self._primary_source][0]["image.data"]
            data_2 = sources[self._secondary_source][0]["image.data"]
        if sources[self._primary_source][0].has("image.mask"):
            mask_1 = sources[self._primary_source][0]["image.mask"]
            mask_2 = sources[self._secondary_source][0]["image.mask"]
        else:
            mask_1 = False
            mask_2 = False

        # streak preview is in and should be put back into ImageData
        wrap_in_imagedata = isinstance(data_1, ImageData)
        if wrap_in_imagedata:
            data_1 = data_1.getData()
            data_2 = data_2.getData()
            mask_1 = mask_1.getData()
            mask_2 = mask_2.getData()

        image_1 = np.ma.masked_array(data=data_1, mask=mask_1)
        image_2 = np.ma.masked_array(data=data_2, mask=mask_2)

        # now to figure out the interleaving
        axis = 0 if data_1.ndim == 1 else 1
        out_shape = utils.interleaving_buffer_shape(data_1.shape, 2, axis)
        if self._interleaving_buffer.shape != out_shape:
            self._interleaving_buffer = np.ma.masked_array(
                np.empty(shape=out_shape, dtype=np.float32),
                mask=False,
            )
        utils.set_on_axis(self._interleaving_buffer, image_1, np.index_exp[0::2], axis)
        utils.set_on_axis(self._interleaving_buffer, image_2, np.index_exp[1::2], axis)
        # TODO: replace this part with preview friend
        self._interleaving_buffer.mask |= ~np.isfinite(self._interleaving_buffer.data)

        if self._wrap_in_imagedata != wrap_in_imagedata:
            # we may need to re-inject output channel to satisfy GUI :D
            schema_update = Schema()
            (
                OUTPUT_CHANNEL(schema_update)
                .key("output")
                .dataSchema(
                    schemas.preview_schema(wrap_image_in_imagedata=wrap_in_imagedata)
                )
                .commit(),
            )
            self.updateSchema(schema_update)
            self.output = self._ss.getOutputChannel("output")
            self._wrap_in_imagedata = wrap_in_imagedata

        meta = ChannelMetaData(
            f"{self.getInstanceId()}:output",
            Timestamp(Epochstamp(), Trainstamp(train_id)),
        )
        if wrap_in_imagedata:
            self.output.write(
                Hash(
                    "image.data",
                    ImageData(
                        self._interleaving_buffer.data,
                        Dims(*self._interleaving_buffer.shape),
                        Encoding.GRAY,
                        bitsPerPixel=32,
                    ),
                    "image.mask",
                    ImageData(
                        self._interleaving_buffer.mask,
                        Dims(*self._interleaving_buffer.shape),
                        Encoding.GRAY,
                        bitsPerPixel=32,
                    ),
                ),
                meta,
                copyAllData=False,
            )
        else:
            self.output.write(
                Hash(
                    "image.data",
                    self._interleaving_buffer.data,
                    "image.mask",
                    self._interleaving_buffer.mask,
                ),
                meta,
                copyAllData=False,
            )
        self.output.update(safeNDArray=True)

        self.info["sent"] += 1
        self.info["trainId"] = train_id
        self.rate_out.update()
