import enum

import karabo.native
import karathon
from karabo.common.scenemodel.api import (
    CheckBoxModel,
    ComboBoxModel,
    DeviceSceneLinkModel,
    DetectorGraphModel,
    DisplayCommandModel,
    DisplayLabelModel,
    DisplayListModel,
    DisplayStateColorModel,
    DisplayTextLogModel,
    DoubleLineEditModel,
    EvaluatorModel,
    IntLineEditModel,
    LabelModel,
    LampModel,
    LineEditModel,
    LineModel,
    NDArrayGraphModel,
    RectangleModel,
    SceneModel,
    SceneTargetWindow,
    TableElementModel,
    TrendGraphModel,
    UnknownWidgetDataModel,
    VectorXYGraphModel,
    WebCamGraphModel,
    WebLinkModel,
    write_scene,
)
import natsort


# section: common setup


BASE_INC = 25
NARROW_INC = 20
PADDING = 5


def DisplayRoundedFloat(*args, decimals=2, **kwargs):
    # note: naive subclass breaks as registry looks for writer based on exact class
    return EvaluatorModel(*args, expression=f"f'{{x:.{decimals}f}}'", **kwargs)


_type_to_display_model = {
    "BOOL": CheckBoxModel,
    "DOUBLE": DisplayRoundedFloat,
    "FLOAT": DisplayRoundedFloat,
    "STRING": DisplayLabelModel,
    "UINT32": DisplayLabelModel,
    "VECTOR_UINT32": DisplayListModel,
}
_type_to_line_editable = {
    "BOOL": (CheckBoxModel, {"klass": "EditableCheckBox"}),
    "DOUBLE": (DoubleLineEditModel, {}),
    "FLOAT": (DoubleLineEditModel, {}),
    "INT32": (IntLineEditModel, {}),
    "UINT32": (IntLineEditModel, {}),
    "INT64": (IntLineEditModel, {}),
    "UINT64": (IntLineEditModel, {}),
    "STRING": (LineEditModel, {"klass": "EditableLineEdit"}),
}


def safe_render(obj, x, y):
    if hasattr(obj, "render"):
        return obj.render(x, y)
    else:
        obj.x = x
        obj.y = y
        return [obj]


class Align(enum.Enum):
    CENTER = enum.auto()
    TOP = enum.auto()
    BOTTOM = enum.auto()
    LEFT = enum.auto()
    RIGHT = enum.auto()


# section: nice component decorators


def titled(title, width=8 * NARROW_INC):
    def actual_decorator(component_class):
        class new_class(component_class):
            def render(self, x, y, *args, **kwargs):
                return [
                    LabelModel(
                        frame_width=1,
                        text=title,
                        width=width,
                        height=NARROW_INC,
                        x=x,
                        y=y,
                    )
                ] + component_class.render(self, x, y + NARROW_INC, *args, **kwargs)

            @property
            def width(self):
                return max(component_class.width.fget(self), width)

            @property
            def height(self):
                return component_class.height.fget(self) + NARROW_INC

        return new_class

    return actual_decorator


def boxed(component_class):
    class new_class(component_class):
        def render(self, x, y, *args, **kwargs):
            return [
                RectangleModel(
                    x=x,
                    y=y,
                    width=component_class.width.fget(self) + 2 * PADDING,
                    height=component_class.height.fget(self) + 2 * PADDING,
                    stroke="#000000",
                )
            ] + component_class.render(self, x + PADDING, y + PADDING, *args, **kwargs)

        @property
        def width(self):
            return component_class.width.fget(self) + 2 * PADDING

        @property
        def height(self):
            return component_class.height.fget(self) + 2 * PADDING

    return new_class


# section: useful layout and utility classes


class Space:
    def __init__(self, width, height):
        self.width = width
        self.height = height

    def render(self, x, y):
        return []


class Hline:
    def __init__(self, width):
        self.width = width
        self.height = 0

    def render(self, x, y):
        return [
            LineModel(
                stroke="#000000",
                x1=x,
                x2=x + self.width,
                y1=y,
                y2=y,
            )
        ]


class Vline:
    def __init__(self, height):
        self.width = 0
        self.height = height

    def render(self, x, y):
        return [
            LineModel(
                stroke="#000000",
                x1=x,
                x2=x,
                y1=y,
                y2=y + self.height,
            )
        ]


def dummy_wrap(model_class):
    class Wrapper:
        def __init__(self, *args, **kwargs):
            self.thing = model_class(*args, **kwargs)

        def render(self, x, y):
            self.thing.x = x
            self.thing.y = y
            return [self.thing]

        @property
        def width(self):
            return self.thing.width

        @property
        def height(self):
            return self.thing.height

    return Wrapper


class HorizontalLayout:
    def __init__(self, *arg_children, children=None, padding=PADDING):
        self.children = list(arg_children)
        if children is not None:
            self.children.extend(children)
        self.padding = padding

    def render(self, x, y, align=Align.TOP):
        if align is not Align.TOP:
            height = self.height
        res = []
        for child in self.children:
            if align is Align.TOP:
                y_ = y
            elif align is Align.CENTER:
                y_ = y + (height - child.height) / 2
            elif align is Align.BOTTOM:
                y_ = y + (height - child.height)
            else:
                raise ValueError(f"Invalid align {align} for HorizontalLayout")
            res.extend(safe_render(child, x, y_))
            x += child.width + self.padding
        return res

    @property
    def width(self):
        if not self.children:
            return 0
        return self.padding * (len(self.children) - 1) + sum(
            c.width for c in self.children
        )

    @property
    def height(self):
        if not self.children:
            return 0
        return max(c.height for c in self.children)


class VerticalLayout:
    def __init__(self, *arg_children, children=None, padding=PADDING):
        self.children = list(arg_children)
        if children is not None:
            self.children.extend(children)
        self.padding = padding

    def render(self, x, y):
        res = []
        for child in self.children:
            res.extend(safe_render(child, x, y))
            y += child.height + self.padding
        return res

    @property
    def width(self):
        if not self.children:
            return 0
        return max(c.width for c in self.children)

    @property
    def height(self):
        if not self.children:
            return 0
        return self.padding * (len(self.children) - 1) + sum(
            c.height for c in self.children
        )


class DisplayAndEditableRow(HorizontalLayout):
    def __init__(
        self,
        device_id,
        schema_hash,
        key_path,
        label_width=7,
        display_width=5,
        edit_width=5,
        height=None,
        size_scale=BASE_INC,
    ):
        super().__init__(padding=0)
        if height is None:
            height = size_scale
        key_attr = schema_hash.getAttributes(key_path)
        label_text = (
            key_attr["displayedName"]
            if "displayedName" in key_attr
            else key_path.split(".")[-1]
        )
        if "valueType" not in key_attr:
            print(f"Key {key_path} on {device_id} had no valueType")
            return
        value_type = key_attr["valueType"]

        self.children.append(
            LabelModel(
                text=label_text,
                width=label_width * size_scale,
                height=height,
            )
        )

        if self.include_display(key_attr):
            if value_type in _type_to_display_model:
                model = _type_to_display_model[value_type]
            else:
                model = DisplayLabelModel
                print(f"Scene generator would like to know more about {value_type}")
            self.children.append(
                model(
                    keys=[f"{device_id}.{key_path}"],
                    width=display_width * size_scale,
                    height=height,
                )
            )

        if self.include_editable(key_attr):
            if "options" in key_attr:
                self.children.append(
                    ComboBoxModel(
                        keys=[f"{device_id}.{key_path}"],
                        width=edit_width * size_scale,
                        height=height,
                        klass="EditableComboBox",
                    )
                )
            elif value_type in _type_to_line_editable:
                line_editable_class, extra_args = _type_to_line_editable[value_type]
                self.children.append(
                    line_editable_class(
                        keys=[f"{device_id}.{key_path}"],
                        width=edit_width * size_scale,
                        height=height,
                        **extra_args,
                    )
                )
            else:
                self.children.append(
                    LabelModel(
                        text=f"Not implemented: editing {value_type} ({key_path})",
                        width=edit_width * size_scale,
                        height=height,
                    )
                )

    def include_display(self, key_attr):
        return True

    def include_editable(self, key_attr):
        return True


class DisplayAndMaybeEditableRow(DisplayAndEditableRow):
    def include_editable(self, key_attr):
        return key_attr["accessMode"] == karabo.native.AccessMode.RECONFIGURABLE.value


class MaybeDisplayMaybeEditableRow(DisplayAndEditableRow):
    # overriding init to unify display_width and edit_width
    def __init__(
        self,
        device_id,
        schema_hash,
        key_path,
        label_width=7,
        display_or_edit_width=5,
        height=None,
        size_scale=BASE_INC,
    ):
        super().__init__(
            device_id,
            schema_hash,
            key_path,
            label_width=label_width,
            display_width=display_or_edit_width,
            edit_width=display_or_edit_width,
            height=height,
            size_scale=size_scale,
        )

    def include_display(self, key_attr):
        return not self.include_editable(key_attr)

    def include_editable(self, key_attr):
        return key_attr["accessMode"] == karabo.native.AccessMode.RECONFIGURABLE.value


class EditableRow(DisplayAndEditableRow):
    # overriding init to get label_width, edit_width without label_width
    def __init__(
        self,
        device_id,
        schema_hash,
        key_path,
        label_width=7,
        edit_width=5,
        height=None,
        size_scale=BASE_INC,
    ):
        super().__init__(
            device_id,
            schema_hash,
            key_path,
            label_width=label_width,
            edit_width=edit_width,
            height=height,
            size_scale=size_scale,
        )

    def include_display(self, key_attr):
        return False

    def include_editable(self, key_attr):
        return True


class DisplayRow(DisplayAndEditableRow):
    def include_display(self, key_attr):
        return True

    def include_editable(self, key_attr):
        return False


# section: specific handcrafted components for device classes


@titled("Found constants", width=6 * NARROW_INC)
@boxed
class FoundConstantsColumn(VerticalLayout):
    def __init__(
        self,
        device_id,
        schema_hash,
        prefix="foundConstants",
    ):
        super().__init__(padding=0)
        for constant_name in schema_hash.get(prefix).getKeys():
            constant_row = HorizontalLayout(
                LabelModel(
                    text=constant_name,
                    width=6 * NARROW_INC,
                    height=NARROW_INC,
                ),
                LampModel(
                    width=NARROW_INC,
                    height=NARROW_INC,
                    keys=[f"{device_id}.{prefix}.{constant_name}.state"],
                ),
                DisplayLabelModel(
                    keys=[f"{device_id}.{prefix}.{constant_name}.beginValidityAt"],
                    width=8 * BASE_INC,
                    height=BASE_INC,
                ),
                DisplayLabelModel(
                    keys=[f"{device_id}.{prefix}.{constant_name}.constantVersionId"],
                    width=3 * BASE_INC,
                    height=BASE_INC,
                ),
                padding=0,
            )
            self.children.append(constant_row)
        self.children.append(
            DeviceSceneLinkModel(
                text="Overrides",
                keys=[f"{device_id}.availableScenes"],
                target="constant_overrides",
                target_window=SceneTargetWindow.Dialog,
                width=8 * BASE_INC,
                height=BASE_INC,
            )
        )


class ConstantLoadedAmpeln(HorizontalLayout):
    def __init__(self, device_id, schema_hash, prefix="foundConstants"):
        super().__init__(padding=0)
        self.children.extend(
            [
                LampModel(
                    keys=[f"{device_id}.{prefix}.{key}.state"],
                    width=BASE_INC,
                    height=BASE_INC,
                )
                for key in schema_hash.get(prefix).getKeys()
            ]
        )


@titled("Manager status", width=6 * NARROW_INC)
@boxed
class ManagerDeviceStatus(VerticalLayout):
    def __init__(self, device_id):
        super().__init__(padding=0)
        name = DisplayLabelModel(
            keys=[f"{device_id}.deviceId"],
            width=14 * BASE_INC,
            height=BASE_INC,
        )
        state = DisplayStateColorModel(
            show_string=True,
            keys=[f"{device_id}.state"],
            width=7 * BASE_INC,
            height=BASE_INC,
        )
        restart_button = DisplayCommandModel(
            keys=[f"{device_id}.restartServers"],
            width=7 * BASE_INC,
            height=BASE_INC,
        )
        instantiate_button = DisplayCommandModel(
            keys=[f"{device_id}.startInstantiate"],
            width=7 * BASE_INC,
            height=BASE_INC,
        )
        status_log = DisplayTextLogModel(
            keys=[f"{device_id}.status"],
            width=14 * BASE_INC,
            height=14 * BASE_INC,
        )
        managed_properties_link = DeviceSceneLinkModel(
            text="All managed properties",
            keys=[f"{device_id}.availableScenes"],
            target="browse_schema",
            target_window=SceneTargetWindow.Dialog,
            width=7 * BASE_INC,
            height=BASE_INC,
        )
        docs_link = WebLinkModel(
            text="Documentation",
            width=7 * BASE_INC,
            height=BASE_INC,
            target="https://rtd.xfel.eu/docs/calng/en/latest/devices/#calibration-manager",
        )
        self.children.extend(
            [
                name,
                HorizontalLayout(
                    state,
                    docs_link,
                    padding=0,
                ),
                HorizontalLayout(
                    restart_button,
                    instantiate_button,
                    padding=0,
                ),
                HorizontalLayout(
                    managed_properties_link,
                    padding=0,
                ),
                status_log,
            ]
        )


@titled("Device status", width=6 * NARROW_INC)
@boxed
class ProcessingDeviceStatus(VerticalLayout):
    def __init__(self, device_id, schema_hash):
        super().__init__(padding=0)
        name = DisplayLabelModel(
            keys=[f"{device_id}.deviceId"],
            width=14 * BASE_INC,
            height=BASE_INC,
        )
        state = DisplayStateColorModel(
            show_string=True,
            keys=[f"{device_id}.state"],
            width=7 * BASE_INC,
            height=BASE_INC,
        )
        rate = DisplayRoundedFloat(
            keys=[f"{device_id}.performance.rate"],
            width=7 * BASE_INC,
            height=BASE_INC,
        )
        processing_time = DisplayRoundedFloat(
            keys=[f"{device_id}.performance.processingTime"],
            width=7 * BASE_INC,
            height=BASE_INC,
        )
        tid = DisplayLabelModel(
            keys=[f"{device_id}.trainId"],
            width=7 * BASE_INC,
            height=BASE_INC,
        )
        status_log = DisplayTextLogModel(
            keys=[f"{device_id}.status"],
            width=14 * BASE_INC,
            height=14 * BASE_INC,
        )
        self.children.extend(
            [
                name,
                HorizontalLayout(
                    state,
                    tid,
                    padding=0,
                ),
                HorizontalLayout(
                    rate,
                    processing_time,
                    padding=0,
                ),
            ]
        )
        if schema_hash.has("warningLamps"):
            self.children.append(
                VerticalLayout(
                    children=[
                        HorizontalLayout(
                            LampModel(
                                keys=[f"{device_id}.{warning_lamp}"],
                                width=BASE_INC,
                                height=BASE_INC,
                            ),
                            LabelModel(
                                text=warning_lamp,
                                width=8 * BASE_INC,
                                height=BASE_INC,
                            ),
                        )
                        for warning_lamp in schema_hash.getAttribute(
                            "warningLamps", "defaultValue"
                        )
                    ],
                    padding=0,
                )
            )
        self.children.append(status_log)


class CompactCorrectionDeviceOverview(HorizontalLayout):
    def __init__(self, device_id, schema_hash):
        super().__init__(padding=0)
        self.children.extend(
            [
                DeviceSceneLinkModel(
                    text=device_id.split("/")[-1],
                    keys=[f"{device_id}.availableScenes"],
                    target="overview",
                    target_window=SceneTargetWindow.Dialog,
                    width=5 * BASE_INC,
                    height=BASE_INC,
                ),
                DisplayStateColorModel(
                    show_string=True,
                    keys=[f"{device_id}.state"],
                    width=5 * BASE_INC,
                    height=BASE_INC,
                ),
            ]
        )
        for warning_lamp in schema_hash.getAttribute("warningLamps", "defaultValue"):
            self.children.append(
                LampModel(
                    keys=[f"{device_id}.{warning_lamp}"],
                    width=BASE_INC,
                    height=BASE_INC,
                )
            )
        self.children.extend(
            [
                DisplayRoundedFloat(
                    keys=[f"{device_id}.performance.rate"],
                    width=4 * BASE_INC,
                    height=BASE_INC,
                ),
                DisplayLabelModel(
                    keys=[f"{device_id}.trainId"],
                    width=4 * BASE_INC,
                    height=BASE_INC,
                ),
                ConstantLoadedAmpeln(device_id, schema_hash),
            ]
        )


class CompactDaqOverview(HorizontalLayout):
    def __init__(self, device_id):
        super().__init__(padding=0)
        self.children.extend(
            [
                DisplayLabelModel(
                    keys=[f"{device_id}.deviceId"],
                    width=8 * BASE_INC,
                    height=BASE_INC,
                    font_size=9,
                ),
                DisplayLabelModel(
                    keys=[f"{device_id}.DataDispatcher.trainStride"],
                    width=4 * BASE_INC,
                    height=BASE_INC,
                ),
            ]
        )


@boxed
class CompactDeviceLinkList(VerticalLayout):
    def __init__(self, device_ids):
        super().__init__()
        self.children.extend(
            [
                HorizontalLayout(
                    DeviceSceneLinkModel(
                        text=device_id.split("/")[-1],
                        keys=[f"{device_id}.availableScenes"],
                        target="",
                        target_window=SceneTargetWindow.Dialog,
                        width=7 * BASE_INC,
                        height=BASE_INC,
                    ),
                    DisplayStateColorModel(
                        show_string=True,
                        keys=[f"{device_id}.state"],
                        width=5 * BASE_INC,
                        height=BASE_INC,
                    ),
                    padding=0,
                )
                for device_id in device_ids
            ]
        )


@titled("Assembler status", width=8 * NARROW_INC)
@boxed
class AssemblerDeviceStatus(VerticalLayout):
    def __init__(self, device_id):
        super().__init__(padding=0)
        name = DisplayLabelModel(
            keys=[f"{device_id}.deviceId"],
            width=14 * BASE_INC,
            height=BASE_INC,
        )
        state = DisplayStateColorModel(
            show_string=True,
            keys=[f"{device_id}.state"],
            width=7 * BASE_INC,
            height=BASE_INC,
        )
        train_id = DisplayLabelModel(
            keys=[f"{device_id}.trainId"],
            width=7 * BASE_INC,
            height=BASE_INC,
        )
        rate = DisplayRoundedFloat(
            keys=[f"{device_id}.outProcessing"],
            width=7 * BASE_INC,
            height=BASE_INC,
        )
        tof = DisplayRoundedFloat(
            keys=[f"{device_id}.timeOfFlight"],
            width=7 * BASE_INC,
            height=BASE_INC,
        )
        self.children.extend(
            [
                name,
                HorizontalLayout(
                    state,
                    train_id,
                    padding=0,
                ),
                HorizontalLayout(
                    rate,
                    tof,
                    padding=0,
                ),
                DeviceSceneLinkModel(
                    text="I'm actually a TrainMatcher",
                    keys=[f"{device_id}.availableScenes"],
                    target="trainMatcherScene",
                    target_window=SceneTargetWindow.Dialog,
                    width=14 * BASE_INC,
                    height=BASE_INC,
                ),
                LabelModel(
                    text="My geometry device:",
                    width=14 * BASE_INC,
                    height=BASE_INC,
                ),
                DisplayLabelModel(
                    keys=[f"{device_id}.geometryDevice"],
                    width=14 * BASE_INC,
                    height=BASE_INC,
                ),  # TODO: some day, get dynamic link to this friend
            ]
        )


@titled("Manual geometry settings")
@boxed
class ManualQuadrantGeometrySettings(VerticalLayout):
    def __init__(self, device_id):
        super().__init__(padding=0)
        self.children.append(
            HorizontalLayout(
                Space(width=3 * BASE_INC, height=BASE_INC),
                LabelModel(text="x", width=4 * BASE_INC, height=BASE_INC),
                LabelModel(text="y", width=4 * BASE_INC, height=BASE_INC),
            )
        )
        self.children.extend(
            [
                HorizontalLayout(
                    LabelModel(text=f"{thing}", width=3 * BASE_INC, height=BASE_INC),
                    DoubleLineEditModel(
                        keys=[f"{device_id}.quadrantCorners.{thing}.x"],
                        width=4 * BASE_INC,
                        height=BASE_INC,
                    ),
                    DoubleLineEditModel(
                        keys=[f"{device_id}.quadrantCorners.{thing}.y"],
                        width=4 * BASE_INC,
                        height=BASE_INC,
                    ),
                )
                for thing in ("Q1", "Q2", "Q3", "Q4", "offset")
            ]
        )
        self.children.append(
            DisplayCommandModel(
                keys=[f"{device_id}.quadrantCorners.setManual"],
                width=6 * BASE_INC,
                height=BASE_INC,
            ),
        )


@titled("Manual geometry settings")
@boxed
class ManualOriginGeometrySettings(VerticalLayout):
    def __init__(self, device_id):
        super().__init__(padding=0)
        self.children.append(
            HorizontalLayout(
                DoubleLineEditModel(
                    keys=[f"{device_id}.origin.x"],
                    width=4 * BASE_INC,
                    height=BASE_INC,
                ),
                DoubleLineEditModel(
                    keys=[f"{device_id}.origin.y"],
                    width=4 * BASE_INC,
                    height=BASE_INC,
                ),
            )
        )
        self.children.append(
            DisplayCommandModel(
                keys=[f"{device_id}.origin.setManual"],
                width=6 * BASE_INC,
                height=BASE_INC,
            ),
        )


@titled("Manual geometry settings")
@boxed
class RelativeGeometrySettings(VerticalLayout):
    def __init__(self, device_id):
        super().__init__(padding=0)
        self.children.extend(
            [
                HorizontalLayout(
                    LabelModel(text="Gap", width=4 * BASE_INC, height=BASE_INC),
                    DoubleLineEditModel(
                        keys=[f"{device_id}.manualSetting.gap"],
                        width=4 * BASE_INC,
                        height=BASE_INC,
                    ),
                ),
                HorizontalLayout(
                    LabelModel(
                        text="Module offsets", width=6 * BASE_INC, height=BASE_INC
                    ),
                    LabelModel(text="x", width=4 * BASE_INC, height=BASE_INC),
                    LabelModel(text="y", width=4 * BASE_INC, height=BASE_INC),
                    LabelModel(text="z", width=4 * BASE_INC, height=BASE_INC),
                ),
                HorizontalLayout(
                    LabelModel(text="Top", width=6 * BASE_INC, height=BASE_INC),
                    DoubleLineEditModel(
                        keys=[f"{device_id}.manualSetting.topOffset.x"],
                        width=4 * BASE_INC,
                        height=BASE_INC,
                    ),
                    DoubleLineEditModel(
                        keys=[f"{device_id}.manualSetting.topOffset.y"],
                        width=4 * BASE_INC,
                        height=BASE_INC,
                    ),
                    DoubleLineEditModel(
                        keys=[f"{device_id}.manualSetting.topOffset.z"],
                        width=4 * BASE_INC,
                        height=BASE_INC,
                    ),
                ),
                HorizontalLayout(
                    LabelModel(text="Bottom", width=6 * BASE_INC, height=BASE_INC),
                    DoubleLineEditModel(
                        keys=[f"{device_id}.manualSetting.bottomOffset.x"],
                        width=4 * BASE_INC,
                        height=BASE_INC,
                    ),
                    DoubleLineEditModel(
                        keys=[f"{device_id}.manualSetting.bottomOffset.y"],
                        width=4 * BASE_INC,
                        height=BASE_INC,
                    ),
                    DoubleLineEditModel(
                        keys=[f"{device_id}.manualSetting.bottomOffset.z"],
                        width=4 * BASE_INC,
                        height=BASE_INC,
                    ),
                ),
                DisplayCommandModel(
                    keys=[f"{device_id}.manualSetting.setManual"],
                    width=6 * BASE_INC,
                    height=BASE_INC,
                ),
            ]
        )


@titled("Manual geometry settings")
@boxed
class ManualModulesGeometrySettings(VerticalLayout):
    def __init__(self, device_id):
        super().__init__(padding=0)
        self.children.append(
            TableElementModel(
                keys=[f"{device_id}.moduleList.modules"],
                klass="EditableTableElement",
                width=14 * BASE_INC,
                height=10 * BASE_INC,
            )
        )
        self.children.append(
            HorizontalLayout(
                LabelModel(text="Offset", width=3 * BASE_INC, height=BASE_INC),
                DoubleLineEditModel(
                    keys=[f"{device_id}.moduleList.offset.x"],
                    width=4 * BASE_INC,
                    height=BASE_INC,
                ),
                DoubleLineEditModel(
                    keys=[f"{device_id}.moduleList.offset.y"],
                    width=4 * BASE_INC,
                    height=BASE_INC,
                ),
            )
        )
        self.children.append(
            DisplayCommandModel(
                keys=[f"{device_id}.moduleList.setManual"],
                width=6 * BASE_INC,
                height=BASE_INC,
            ),
        )


@titled("Tweak current geometry")
@boxed
class TweakCurrentGeometry(VerticalLayout):
    def __init__(self, device_id):
        super().__init__(padding=0)
        self.children.append(
            HorizontalLayout(
                LabelModel(text="Offset", width=3 * BASE_INC, height=BASE_INC),
                DoubleLineEditModel(
                    keys=[f"{device_id}.tweakGeometry.offset.x"],
                    width=4 * BASE_INC,
                    height=BASE_INC,
                ),
                DoubleLineEditModel(
                    keys=[f"{device_id}.tweakGeometry.offset.y"],
                    width=4 * BASE_INC,
                    height=BASE_INC,
                ),
            ),
        )
        self.children.append(
            DisplayCommandModel(
                keys=[f"{device_id}.tweakGeometry.add"],
                width=6 * BASE_INC,
                height=BASE_INC,
            ),
        )
        self.children.append(
            HorizontalLayout(
                DisplayLabelModel(
                    keys=[f"{device_id}.tweakGeometry.undoLength"],
                    width=2 * BASE_INC,
                    height=BASE_INC,
                    font_size=9,
                ),
                DisplayCommandModel(
                    keys=[f"{device_id}.tweakGeometry.undo"],
                    width=4 * BASE_INC,
                    height=BASE_INC,
                ),
                DisplayLabelModel(
                    keys=[f"{device_id}.tweakGeometry.redoLength"],
                    width=2 * BASE_INC,
                    height=BASE_INC,
                    font_size=9,
                ),
                DisplayCommandModel(
                    keys=[f"{device_id}.tweakGeometry.redo"],
                    width=4 * BASE_INC,
                    height=BASE_INC,
                ),
            ),
        )


@titled("Geometry preview")
@boxed
class GeometryPreview(VerticalLayout):
    def __init__(self, device_id):
        super().__init__(padding=0)
        self.children.append(
            DisplayCommandModel(
                keys=[f"{device_id}.updatePreview"],
                width=6 * BASE_INC,
                height=BASE_INC,
            ),
        )
        self.children.append(
            WebCamGraphModel(
                keys=[f"{device_id}.geometryPreview"],
                width=30 * BASE_INC,
                height=30 * BASE_INC,
                x=PADDING,
                y=PADDING,
            )
        )


@titled("Geometry from file")
@boxed
class GeometryFromFileSettings(VerticalLayout):
    def __init__(self, device_id, schema_hash):
        super().__init__(padding=0)
        self.children.extend(
            [
                EditableRow(device_id, schema_hash, "geometryFile.filePath", 4, 8),
                EditableRow(device_id, schema_hash, "geometryFile.fileType", 4, 8),
            ]
        )
        self.children.append(
            HorizontalLayout(
                LabelModel(text="Offset", width=3 * BASE_INC, height=BASE_INC),
                DoubleLineEditModel(
                    keys=[f"{device_id}.geometryFile.offset.x"],
                    width=4 * BASE_INC,
                    height=BASE_INC,
                ),
                DoubleLineEditModel(
                    keys=[f"{device_id}.geometryFile.offset.y"],
                    width=4 * BASE_INC,
                    height=BASE_INC,
                ),
            )
        )
        self.children.append(
            EditableRow(device_id, schema_hash, "geometryFile.updateManualOnLoad", 6, 2)
        )
        self.children.append(
            DisplayCommandModel(
                keys=[f"{device_id}.geometryFile.loadFromFile"],
                width=6 * BASE_INC,
                height=BASE_INC,
            ),
        )


@titled("Stats")
@boxed
class StatsBox(HorizontalLayout):
    def __init__(self, device_id, schema_hash):
        super().__init__()
        self.children.extend(
            [
                DisplayRow(device_id, schema_hash, "rate", 5, 3),
                DisplayRow(device_id, schema_hash, "numPixelsIncluded", 5, 3),
            ]
        )


@titled("ROI selection")
@boxed
class RoiSelection(VerticalLayout):
    def __init__(self, image_path, roi_paths):
        super().__init__()
        self.children.append(
            UnknownWidgetDataModel(
                keys=([image_path] + list(roi_paths)),
                klass="RectRoiGraph",
                height=25 * BASE_INC,
                width=30 * BASE_INC,
            )
        )


@titled("ROI preview")
@boxed
class RoiBox(VerticalLayout):
    def __init__(self, device_id):
        super().__init__()
        self.children.append(
            DetectorGraphModel(
                keys=[f"{device_id}.output.schema.zoomImage"],
                height=17 * BASE_INC,
                width=20 * BASE_INC,
            ),
        )


@titled("Preview settings")
@boxed
class PreviewSettings(HorizontalLayout):
    def __init__(self, device_id, schema_hash, node_name="preview", extras=None):
        super().__init__()
        self.children.extend(
            [
                VerticalLayout(
                    EditableRow(
                        device_id,
                        schema_hash,
                        f"{node_name}.replaceNanWith",
                        6,
                        4,
                    ),
                    HorizontalLayout(
                        EditableRow(
                            device_id,
                            schema_hash,
                            f"{node_name}.flipSS",
                            3,
                            2,
                        ),
                        EditableRow(
                            device_id,
                            schema_hash,
                            f"{node_name}.flipFS",
                            3,
                            2,
                        ),
                        padding=0,
                    ),
                ),
                Vline(height=3 * BASE_INC),
                VerticalLayout(
                    LabelModel(
                        text="Image downsampling",
                        width=8 * BASE_INC,
                        height=BASE_INC,
                    ),
                    EditableRow(
                        device_id,
                        schema_hash,
                        f"{node_name}.downsamplingFactor",
                        5,
                        5,
                    ),
                    EditableRow(
                        device_id,
                        schema_hash,
                        f"{node_name}.downsamplingFunction",
                        5,
                        5,
                    ),
                ),
            ]
        )
        if extras is not None:
            self.children.append(Vline(height=3 * BASE_INC))
            self.children.extend(extras)


@titled("Histogram settings")
@boxed
class HistogramSettings(HorizontalLayout):
    def __init__(self, device_id, schema_hash):
        super().__init__()
        self.children.extend(
            [
                VerticalLayout(
                    LabelModel(text="Bins", width=3 * BASE_INC, height=BASE_INC),
                    DisplayCommandModel(
                        keys=[f"{device_id}.histogram.resetBins"],
                        width=4 * BASE_INC,
                        height=BASE_INC,
                    ),
                    EditableRow(
                        device_id, schema_hash, "histogram.resetBinsOnRoiChange", 5, 1
                    ),
                    EditableRow(device_id, schema_hash, "histogram.numBins", 3, 3),
                    padding=0,
                ),
                VerticalLayout(
                    LabelModel(text="Range", width=4 * BASE_INC, height=BASE_INC),
                    EditableRow(device_id, schema_hash, "histogram.rangeMin", 4, 3),
                    EditableRow(device_id, schema_hash, "histogram.rangeMax", 4, 3),
                    EditableRow(
                        device_id,
                        schema_hash,
                        "histogram.automaticallyExpandRange",
                        4,
                        3,
                    ),
                    padding=0,
                ),
                VerticalLayout(
                    LabelModel(text="Averaging", width=4 * BASE_INC, height=BASE_INC),
                    EditableRow(
                        device_id, schema_hash, "histogram.rollingWindowSize", 4, 3
                    ),
                    padding=0,
                ),
            ]
        )


class PreviewDisplayArea(VerticalLayout):
    def __init__(
        self,
        device_id,
        schema_hash,
        channel_name,
        data_width=None,
        mask_width=None,
        data_height=None,
        mask_height=None,
    ):
        super().__init__()
        try:
            class_id = schema_hash.getAttribute(
                f"{channel_name}.schema.image.data", "classId"
            )
        except RuntimeError:
            warning = f"{channel_name} in schema of {device_id} missing classId"
            print(warning)
            self.children.append(
                LabelModel(text=warning, width=10 * BASE_INC, height=BASE_INC)
            )
            return
        if class_id == "ImageData":
            self.children.extend(
                [
                    DetectorGraphModel(
                        keys=[f"{device_id}.{channel_name}.schema.image.data"],
                        colormap="viridis",
                        width=60 * BASE_INC if data_width is None else data_width,
                        height=18 * BASE_INC if data_height is None else data_height,
                    ),
                    DetectorGraphModel(
                        keys=[f"{device_id}.{channel_name}.schema.image.mask"],
                        colormap="viridis",
                        width=60 * BASE_INC if mask_width is None else mask_width,
                        height=18 * BASE_INC if mask_height is None else mask_height,
                    ),
                ]
            )
        else:
            self.children.extend(
                [
                    NDArrayGraphModel(
                        keys=[f"{device_id}.{channel_name}.schema.image.data"],
                        width=60 * BASE_INC if data_width is None else data_width,
                        height=18 * BASE_INC if data_height is None else data_height,
                    ),
                    NDArrayGraphModel(
                        keys=[f"{device_id}.{channel_name}.schema.image.mask"],
                        width=60 * BASE_INC if mask_width is None else mask_width,
                        height=12 * BASE_INC if mask_height is None else mask_height,
                    ),
                ]
            )


# section: generating actual scenes


def schema_to_hash(schema):
    if isinstance(schema, (karathon.Hash, karabo.native.Hash)):
        return schema
    elif isinstance(schema, karathon.Schema):
        return schema.getParameterHash()
    else:
        return schema.hash


def scene_generator(fun):
    # TODO: pretty decorator
    def aux(*args, **kwargs):
        content = fun(*args, **kwargs)

        scene = SceneModel(
            children=content.render(PADDING, PADDING),
            width=content.width + 2 * PADDING,
            height=content.height + 2 * PADDING,
        )
        return write_scene(scene)

    return aux


@scene_generator
def correction_device_overview(device_id, schema):
    schema_hash = schema_to_hash(schema)
    main_overview = HorizontalLayout(
        ProcessingDeviceStatus(device_id, schema_hash),
        VerticalLayout(
            recursive_editable(
                device_id,
                schema_hash,
                "constantParameters",
            ),
            DisplayCommandModel(
                keys=[f"{device_id}.loadMostRecentConstants"],
                width=10 * BASE_INC,
                height=BASE_INC,
            ),
        ),
        FoundConstantsColumn(device_id, schema_hash),
        recursive_editable(
            device_id,
            schema_hash,
            "corrections",
            max_depth=2,
        ),
    )
    return VerticalLayout(
        main_overview,
        LabelModel(
            text="Preview (corrected):",
            width=20 * BASE_INC,
            height=BASE_INC,
        ),
        PreviewDisplayArea(device_id, schema_hash, "preview.outputCorrected"),
        *(
            DeviceSceneLinkModel(
                text=f"Preview: {channel}",
                keys=[f"{device_id}.availableScenes"],
                target=f"preview:preview.{channel}",
                target_window=SceneTargetWindow.Dialog,
                width=16 * BASE_INC,
                height=BASE_INC,
            )
            for channel in schema_hash.get("preview").getKeys()
            if schema_hash.hasAttribute(f"preview.{channel}", "classId")
            and schema_hash.getAttribute(f"preview.{channel}", "classId")
            == "OutputChannel"
        ),
    )


@scene_generator
def lpdmini_splitter_overview(device_id, schema):
    schema_hash = schema_to_hash(schema)
    return VerticalLayout(
        ProcessingDeviceStatus(device_id, schema_hash),
        DisplayRow(device_id, schema_hash, "outputDataShape", 6, 8),
    )


@scene_generator
def correction_device_preview(device_id, schema, preview_channel):
    schema_hash = schema_to_hash(schema)
    return VerticalLayout(
        LabelModel(
            text=f"Preview: {preview_channel}",
            width=10 * BASE_INC,
            height=BASE_INC,
        ),
        PreviewSettings(
            device_id,
            schema_hash,
            extras=[
                VerticalLayout(
                    EditableRow(device_id, schema_hash, "preview.index", 8, 4),
                    EditableRow(device_id, schema_hash, "preview.selectionMode", 8, 4),
                )
            ],
        ),
        PreviewDisplayArea(device_id, schema_hash, preview_channel),
    )


@scene_generator
def correction_device_constant_overrides(device_id, schema, prefix="foundConstants"):
    schema_hash = schema_to_hash(schema)

    return HorizontalLayout(
        children=[
            VerticalLayout(
                VerticalLayout(
                    LabelModel(text="constant", width=6 * BASE_INC, height=BASE_INC),
                    LabelModel(text="timestamp", width=6 * BASE_INC, height=BASE_INC),
                    padding=0,
                ),
                VerticalLayout(
                    LabelModel(
                        text="constant version ID",
                        width=6 * BASE_INC,
                        height=BASE_INC,
                    ),
                    Space(width=6 * BASE_INC, height=BASE_INC),
                    padding=0,
                ),
                VerticalLayout(
                    LabelModel(
                        text="data file name", width=6 * BASE_INC, height=BASE_INC
                    ),
                    LabelModel(
                        text="dataset name", width=6 * BASE_INC, height=BASE_INC
                    ),
                    Space(width=6 * BASE_INC, height=BASE_INC),
                    padding=0,
                ),
            )
        ]
        + [
            VerticalLayout(
                VerticalLayout(
                    HorizontalLayout(
                        LampModel(
                            keys=[f"{device_id}.{prefix}.{constant}.state"],
                            width=BASE_INC,
                            height=BASE_INC,
                        ),
                        LabelModel(
                            text=constant,
                            width=7 * BASE_INC,
                            height=BASE_INC,
                        ),
                    ),
                    DisplayLabelModel(
                        keys=[f"{device_id}.{prefix}.{constant}.beginValidityAt"],
                        width=8 * BASE_INC,
                        height=BASE_INC,
                        font_size=9,
                    ),
                    padding=0,
                ),
                DisplayCommandModel(
                    keys=[f"{device_id}.{prefix}.{constant}.loadMostRecent"],
                    width=8 * BASE_INC,
                    height=BASE_INC,
                ),
                VerticalLayout(
                    LineEditModel(
                        klass="EditableLineEdit",
                        keys=[f"{device_id}.{prefix}.{constant}.constantVersionId"],
                        width=8 * BASE_INC,
                        height=BASE_INC,
                    ),
                    DisplayCommandModel(
                        keys=[
                            f"{device_id}.{prefix}.{constant}.overrideConstantFromVersion"
                        ],
                        width=8 * BASE_INC,
                        height=BASE_INC,
                    ),
                    padding=0,
                ),
                VerticalLayout(
                    LineEditModel(
                        klass="EditableLineEdit",
                        keys=[f"{device_id}.{prefix}.{constant}.dataFilePath"],
                        width=8 * BASE_INC,
                        height=BASE_INC,
                    ),
                    LineEditModel(
                        klass="EditableLineEdit",
                        keys=[f"{device_id}.{prefix}.{constant}.dataSetName"],
                        width=8 * BASE_INC,
                        height=BASE_INC,
                    ),
                    DisplayCommandModel(
                        keys=[
                            f"{device_id}.{prefix}.{constant}.overrideConstantFromFile"
                        ],
                        width=8 * BASE_INC,
                        height=BASE_INC,
                    ),
                    padding=0,
                ),
            )
            for constant in schema_hash.get(prefix).getKeys()
        ]
    )


@scene_generator
def roitool_overview(device_id, schema):
    schema_hash = schema_to_hash(schema)
    return VerticalLayout(
        HorizontalLayout(
            RoiSelection(
                f"{device_id}.output.schema.fullImage", [f"{device_id}.roi"]
            ),
            VerticalLayout(
                RoiBox(device_id),
                StatsBox(device_id, schema_hash),
                HistogramSettings(device_id, schema_hash),
            ),
        ),
        VectorXYGraphModel(
            keys=[
                f"{device_id}.output.schema.histogram.x",
                f"{device_id}.output.schema.histogram.y",
                f"{device_id}.output.schema.histogram.yMean",
            ],
            height=16 * BASE_INC,
            width=52 * BASE_INC,
        ),
    )


@scene_generator
def manager_device_overview(
    manager_device_id,
    manager_device_schema,
    correction_device_ids,
    correction_device_schema,
    daq_device_ids,
    domain_device_ids,
):
    mds_hash = schema_to_hash(manager_device_schema)
    cds_hash = schema_to_hash(correction_device_schema)

    data_throttling_children = [
        LabelModel(
            text="Frame filter",
            width=11 * BASE_INC,
            height=BASE_INC,
        ),
        EditableRow(
            manager_device_id,
            mds_hash,
            "managedKeys.frameFilter.type",
            7,
            4,
        ),
        EditableRow(
            manager_device_id,
            mds_hash,
            "managedKeys.frameFilter.spec",
            7,
            4,
        ),
    ]

    if "managedKeys.daqTrainStride" in mds_hash:
        # Only add DAQ train stride if present on the schema, may be
        # disabled on the manager.
        data_throttling_children.insert(
            0,
            EditableRow(
                manager_device_id,
                mds_hash,
                "managedKeys.daqTrainStride",
                7,
                4,
            ),
        )

    return VerticalLayout(
        HorizontalLayout(
            ManagerDeviceStatus(manager_device_id),
            VerticalLayout(
                recursive_editable(
                    manager_device_id,
                    mds_hash,
                    "managedKeys.constantParameters",
                ),
                DisplayCommandModel(
                    keys=[f"{manager_device_id}.managedKeys.loadMostRecentConstants"],
                    width=10 * BASE_INC,
                    height=BASE_INC,
                ),
                recursive_editable(
                    manager_device_id,
                    mds_hash,
                    "managedKeys.preview",
                    max_depth=2,
                ),
                titled("Data throttling")(boxed(VerticalLayout))(
                    children=data_throttling_children,
                    padding=0,
                ),
            ),
            recursive_editable(
                manager_device_id,
                mds_hash,
                "managedKeys.corrections",
                max_depth=2,
            ),
        ),
        HorizontalLayout(
            titled("DAQs", width=8 * NARROW_INC)(boxed(VerticalLayout))(
                children=[Space(width=BASE_INC, height=BASE_INC)] * 2
                + [
                    CompactDaqOverview(device_id)
                    for device_id in natsort.natsorted(daq_device_ids)
                ],
                padding=0,
            ),
            titled("Correction devices", width=8 * NARROW_INC)(boxed(VerticalLayout))(
                children=[
                    HorizontalLayout(
                        VerticalLayout(
                            DeviceSceneLinkModel(
                                text="Performance dashboard",
                                keys=[f"{manager_device_id}.availableScenes"],
                                target="correction_performance_overview",
                                target_window=SceneTargetWindow.Dialog,
                                width=10 * BASE_INC,
                                height=BASE_INC,
                            ),
                            DeviceSceneLinkModel(
                                text="Correction constant overview",
                                keys=[f"{manager_device_id}.availableScenes"],
                                target="correction_constant_overview",
                                target_window=SceneTargetWindow.Dialog,
                                width=10 * BASE_INC,
                                height=BASE_INC,
                            ),
                            padding=0,
                        ),
                        WebLinkModel(
                            text="Documentation",
                            width=6 * BASE_INC,
                            height=BASE_INC,
                            target="https://rtd.xfel.eu/docs/calng/en/latest/devices/#correction-devices",
                        ),
                        padding=0,
                    )
                ]
                + [
                    CompactCorrectionDeviceOverview(device_id, cds_hash)
                    for device_id in natsort.natsorted(correction_device_ids)
                ],
                padding=0,
            ),
            titled("Other devices managed")(CompactDeviceLinkList)(
                natsort.natsorted(
                    set(domain_device_ids)
                    - set(correction_device_ids)
                    - {manager_device_id}
                )
            ),
        ),
    )


@scene_generator
def correction_device_performance_dashboard(correction_device_ids):
    correction_device_ids = natsort.natsorted(correction_device_ids)
    return HorizontalLayout(
        titled("Correction device links")(CompactDeviceLinkList)(correction_device_ids),
        VerticalLayout(
            TrendGraphModel(
                keys=[
                    f"{device_id}.performance.rate"
                    for device_id in correction_device_ids
                ],
                width=30 * BASE_INC,
                height=15 * BASE_INC,
                x_label="timestamp",
                y_label="rate",
                y_units="Hz",
            ),
            TrendGraphModel(
                keys=[
                    f"{device_id}.performance.processingTime"
                    for device_id in correction_device_ids
                ],
                width=30 * BASE_INC,
                height=15 * BASE_INC,
                x_label="timestamp",
                y_label="processing per train",
                y_units="ms",
            ),
            TrendGraphModel(
                keys=[
                    f"{device_id}.performance.ratioOfRecentTrainsReceived"
                    for device_id in correction_device_ids
                ],
                width=30 * BASE_INC,
                height=15 * BASE_INC,
                x_label="timestamp",
                y_label="ratio of trains received",
                y_units="%",
            ),
        ),
    )


@scene_generator
def correction_constant_dashboard(
    correction_device_ids, correction_device_schema, prefix="foundConstants"
):
    correction_device_schema = schema_to_hash(correction_device_schema)
    constant_names = list(correction_device_schema.get(prefix).getKeys())
    correction_device_ids = natsort.natsorted(correction_device_ids)
    return VerticalLayout(
        HorizontalLayout(
            Space(width=6 * BASE_INC, height=BASE_INC),
            *(
                LabelModel(text=constant, width=7 * BASE_INC, height=BASE_INC)
                for constant in constant_names
            ),
            padding=0,
        ),
        *(
            HorizontalLayout(
                DeviceSceneLinkModel(
                    text=device_id.split("/")[-1],
                    keys=[f"{device_id}.availableScenes"],
                    width=6 * BASE_INC,
                    height=BASE_INC,
                ),
                *(
                    DisplayLabelModel(
                        keys=[f"{device_id}.{prefix}.{constant}.beginValidityAt"],
                        width=7 * BASE_INC,
                        height=BASE_INC,
                        font_size=9,
                    )
                    for constant in constant_names
                ),
                padding=0,
            )
            for device_id in correction_device_ids
        ),
    )


@scene_generator
def detector_assembler_overview(device_id, schema):
    schema_hash = schema_to_hash(schema)
    return VerticalLayout(
        HorizontalLayout(
            AssemblerDeviceStatus(device_id),
            PreviewSettings(device_id, schema_hash),
        ),
        PreviewDisplayArea(
            device_id,
            schema_hash,
            "preview.output",
            data_width=40 * BASE_INC,
            data_height=40 * BASE_INC,
            mask_width=40 * BASE_INC,
            mask_height=15 * BASE_INC,
        ),
    )


@scene_generator
def gotthard2_assembler_overview(device_id, schema):
    schema_hash = schema_to_hash(schema)
    return VerticalLayout(
        HorizontalLayout(
            AssemblerDeviceStatus(device_id),
        ),
        PreviewDisplayArea(
            device_id,
            schema_hash,
            "output",
            data_width=40 * BASE_INC,
            data_height=40 * BASE_INC,
            mask_width=40 * BASE_INC,
            mask_height=15 * BASE_INC,
        ),
    )


@scene_generator
def quadrant_geometry_overview(device_id, schema):
    schema_hash = schema_to_hash(schema)
    return VerticalLayout(
        HorizontalLayout(
            ManualQuadrantGeometrySettings(device_id),
            GeometryFromFileSettings(device_id, schema_hash),
            TweakCurrentGeometry(device_id),
        ),
        GeometryPreview(device_id),
    )


@scene_generator
def origin_geometry_overview(device_id, schema):
    # TODO: handle loading
    return VerticalLayout(
        HorizontalLayout(
            ManualOriginGeometrySettings(device_id),
            # GeometryFromFileSettings(device_id, schema_hash),
            # TweakCurrentGeometry(device_id),
        ),
        GeometryPreview(device_id),
    )


@scene_generator
def relative_geometry_overview(device_id, schema):
    return VerticalLayout(
        HorizontalLayout(
            RelativeGeometrySettings(device_id),
            # GeometryFromFileSettings(device_id, schema_hash),
            # TweakCurrentGeometry(device_id),
        ),
        GeometryPreview(device_id),
    )


@scene_generator
def modules_geometry_overview(device_id, schema):
    schema_hash = schema_to_hash(schema)
    return VerticalLayout(
        HorizontalLayout(
            ManualModulesGeometrySettings(device_id),
            GeometryFromFileSettings(device_id, schema_hash),
            TweakCurrentGeometry(device_id),
        ),
        GeometryPreview(device_id),
    )


@scene_generator
def condition_checker_overview(device_id, schema):
    schema_hash = schema_to_hash(schema)
    return VerticalLayout(
        HorizontalLayout(
            DisplayCommandModel(
                keys=[f"{device_id}.checkConditions"],
                width=7 * BASE_INC,
                height=BASE_INC,
            ),
            DisplayCommandModel(
                keys=[f"{device_id}.updateConditions"],
                width=7 * BASE_INC,
                height=BASE_INC,
            ),
            DisplayStateColorModel(
                keys=[f"{device_id}.conditionsMatch"],
                width=7 * BASE_INC,
                height=BASE_INC,
            ),
        ),
        HorizontalLayout(
            DisplayCommandModel(
                keys=[f"{device_id}.startMonitoring"],
                width=7 * BASE_INC,
                height=BASE_INC,
            ),
            DisplayCommandModel(
                keys=[f"{device_id}.stopMonitoring"],
                width=7 * BASE_INC,
                height=BASE_INC,
            ),
        ),
        EditableRow(device_id, schema_hash, "updateManagerOnMonitor", 13, 1),
        EditableRow(device_id, schema_hash, "loadConstantsOnMonitor", 13, 1),
        TableElementModel(
            keys=[f"{device_id}.keyMapping"],
            width=30 * BASE_INC,
            height=20 * BASE_INC,
        ),
    )


# section: here be monsters


def recursive_editable(
    device_id,
    schema_hash,
    prefix,
    depth=1,
    max_depth=3,
    title=None,
    row_class=MaybeDisplayMaybeEditableRow,
):
    schema_hash = schema_to_hash(schema_hash)
    # note: not just using sets because that loses ordering
    node_keys = []
    value_keys = []
    slot_keys = []
    attr = schema_hash.getAttributes(prefix)

    if title is None:
        if "displayedName" in attr:
            title = attr.get("displayedName")
        else:
            title = prefix.split(".")[-1]

    for key in schema_hash.get(prefix).getKeys():
        attrs = schema_hash.getAttributes(f"{prefix}.{key}")
        if attrs.get("nodeType") == karabo.native.NodeType.Node.value:
            if "classId" in attrs and attrs.get("classId") == "Slot":
                slot_keys.append(key)
            else:
                node_keys.append(key)
        else:
            value_keys.append(key)
    res = titled(title)(boxed(VerticalLayout))(
        children=[
            row_class(device_id, schema_hash, f"{prefix}.{key}") for key in value_keys
        ]
        + [
            DisplayCommandModel(
                keys=[f"{device_id}.{prefix}.{key}"],
                width=10 * BASE_INC,
                height=BASE_INC,
            )
            for key in slot_keys
        ],
        padding=0,
    )
    if depth < max_depth:
        res.children.append(
            VerticalLayout(
                children=[
                    recursive_editable(
                        device_id,
                        schema_hash,
                        f"{prefix}.{key}",
                        depth=depth + 1,
                        max_depth=max_depth,
                        row_class=row_class,
                    )
                    for key in node_keys
                ]
            )
        )
    else:
        res.children.extend(
            [
                VerticalLayout(
                    DeviceSceneLinkModel(
                        text=key,
                        keys=[f"{device_id}.availableScenes"],
                        target=f"browse_schema:{prefix}.{key}",
                        target_window=SceneTargetWindow.Dialog,
                        width=5 * BASE_INC,
                        height=BASE_INC,
                    ),
                )
                for key in node_keys
            ]
        )
    return res


@scene_generator
def recursive_subschema_scene(
    device_id,
    device_schema,
    prefix="managedKeys",
):
    mds_hash = schema_to_hash(device_schema)
    return recursive_editable(device_id, mds_hash, prefix)
