import contextlib
import collections
import enum
import functools
import inspect
import itertools
import threading
import time
from timeit import default_timer

import numpy as np


class WarningContextSystem:
    """Helper object to trigger warning lamps based on different warning types in
    contexts. Intended use: something that is checked multiple times and is good or bad
    depending on last check (ex. per train).

    Warning model: a warning is generated within context (see __call_). Each warning
    context handles one schema key and one warning type for that key. One lamp
    aggregates multiple warning types; for each lamp, a set of currently active warnings
    decides the ultimate fate of the lamp.

    If a warning is issued within a context, the corresponding warning type is added to
    the set. If not, then the corresponding warning type is removed from the set."""

    def __init__(self, device, on_success=None, on_error=None):
        self.device = device
        # key -> set of warning types
        self.active_warnings = collections.defaultdict(lambda: set())
        # (key, warn type) -> True or False depending on last context
        self.triggered_warnings = collections.defaultdict(lambda: False)

        # key -> state string
        self.on_success = collections.defaultdict(lambda: "ON")
        if on_success is not None:
            self.on_success.update(on_success)
        self.on_error = collections.defaultdict(lambda: "ERROR")
        if on_error is not None:
            self.on_error.update(on_error)
        # TODO: support mapping different warning types to different Karabo states
        # (for colors per lamp depending on state signifier on active set)

    @contextlib.contextmanager
    def __call__(self, key, warn_type=None):
        """Will return a context handler object for tkey specified schema key and
        warning type. It is your responsibility to ensure that the key is valid. Using
        __call__ on the returned context handler will issue a warning and add the
        warning type to the set of active warnings for the key."""
        warn_fun = functools.partial(self.issue_warning, key, warn_type)
        try:
            yield warn_fun
        except Exception as ex:
            warn_fun(f"Unexpected exception: {ex}")
        finally:
            # warn_fun may have set trigger; we update active warnings out here
            active_set = self.active_warnings[key]
            if self.triggered_warnings[(key, warn_type)]:
                if warn_type not in active_set:
                    self.device.set(key, self.on_error[key])
                    active_set.add(warn_type)
                self.triggered_warnings[(key, warn_type)] = False
            else:
                # warning not triggered, so clear out active set
                if warn_type in active_set:
                    message = (
                        f"{key}: warning lifted!"
                        if warn_type is None
                        else f"{key} > {warn_type.name}: lifted!"
                    )
                    self.device.log.INFO(message)
                    self.device.set("status", message)
                    active_set.discard(warn_type)
                if not active_set:
                    self.device.set(key, self.on_success[key])

    def reset(self):
        self.active_warnings.clear()
        self.triggered_warnings.clear()

    def issue_warning(self, key, warn_type, text):
        self.triggered_warnings[(key, warn_type)] = True
        # already active, so don't print
        if warn_type in self.active_warnings[key]:
            return
        message = (
            f"{key}: {text}"
            if warn_type is None
            else f"{key} > {warn_type.name}: {text}"
        )
        # TODO: make log + "status" behavior customizable
        self.device.log.WARN(message)
        self.device.set("status", message)


class PreviewIndexSelectionMode(enum.Enum):
    FRAME = "frame"
    CELL = "cell"
    PULSE = "pulse"


def pick_frame_index(selection_mode, index, cell_table, pulse_table):
    """When selecting a single frame to preview, an obvious question is whether the
    number the operator provides is a frame index, a cell ID, or a pulse ID. This
    function allows any of the three, translating into frame index.

    Indices below zero are special values and thus returned directly.

    Returns: (frame index, cell ID, pulse ID), optional warning"""

    if index < 0:
        return (index, index, index), None

    warning = None
    selection_mode = PreviewIndexSelectionMode(selection_mode)

    if selection_mode is PreviewIndexSelectionMode.FRAME:
        if index < cell_table.size:
            frame_index = index
        else:
            warning = (
                f"Frame index {index} out of range for cell table of length "
                f"{len(cell_table)}. Will use frame index 0 instead."
            )
            frame_index = 0
    else:
        if selection_mode is PreviewIndexSelectionMode.CELL:
            index_table = cell_table
        else:
            index_table = pulse_table
        found = np.where(index_table == index)[0]
        if len(found) == 1:
            frame_index = found[0]
        elif len(found) == 0:
            frame_index = 0
            warning = (
                f"{selection_mode.value.capitalize()} ID {index} not found in "
                f"{selection_mode.name} table. Will use frame {frame_index}, which "
                f"corresponds to {selection_mode.name} {index_table[0]} instead."
            )
        elif len(found) > 1:
            warning = (
                f"{selection_mode.value.capitalize()} ID {index} was not unique in "
                f"{selection_mode} table. Will use first occurrence out of "
                f"{len(found)} occurrences (frame {frame_index})."
            )

    cell_id = cell_table[frame_index]
    pulse_id = pulse_table[frame_index]
    return (frame_index, cell_id, pulse_id), warning


def threadsafe_cache(fun):
    """This decorator imitates functools.cache, but threadsafer

    With multiple threads hitting a function cached by functools.cache, it is possible
    to trigger recomputation. This decorator adds granular locking: each key in the
    cache (derived from arguments) has its own lock.
    """

    locks = {}
    results = {}
    fun_sig = inspect.signature(fun)

    @functools.wraps(fun)
    def aux(*args, **kwargs):
        bound_args = fun_sig.bind(*args, **kwargs)
        bound_args.apply_defaults()
        key = bound_args.args + tuple(bound_args.kwargs.items())
        if key in results:
            return results[key]
        with locks.setdefault(key, threading.Lock()):
            if key in results:
                # someone else did this - may still be processing
                return results[key]
            else:
                res = fun(*args, **kwargs)
                results[key] = res
                return res

    return aux


@functools.lru_cache()
def transpose_order(axes_in, axes_out):
    """Computes the order of axes_out relative to axes_in for transposition purposes

    Both axes_in and axes_out are assumed to be strings in which each letter represents
    an axis (duck typing accepts: any iterable of hashable elements). They should
    probably be of the same length and have no repetitions, but this is not enforced.
    Off-label use voids warranty.
    """
    axis_order = {axis: index for index, axis in enumerate(axes_in)}
    return tuple(axis_order[axis] for axis in axes_out)


def stacking_buffer_shape(array_shape, stack_num, axis=0):
    """Figures out the shape you would need for np.stack. Think of the axis in terms of
    array after adding additional axis, i.e. the number of axes is len(aray_shape)+1."""
    if axis > len(array_shape) or axis < -len(array_shape) - 1:
        # complain when np.stack would
        raise np.AxisError(
            f"axis {axis} is out of bounds "
            f"for array of dimension {len(array_shape) + 1}"
        )
    if axis < 0:
        axis += len(array_shape) + 1
    return array_shape[:axis] + (stack_num,) + array_shape[axis:]


def interleaving_buffer_shape(array_shape, stack_num, axis):
    """Figures out the shape you would need to interleave stack_num arrays on axis. The
    shape is the same as one would get from np.concatenate."""
    if axis > len(array_shape) - 1 or axis < -len(array_shape):
        raise np.AxisError(
            f"axis {axis} is out of bounds for array shape {array_shape}"
        )
    if axis < 0:
        axis += len(array_shape) + 1
    return (
        array_shape[:axis] + (array_shape[axis] * stack_num,) + array_shape[axis + 1 :]
    )


def set_on_axis(array, vals, index, axis):
    """set_on_axis(A, x, 1, 2) corresponds to A[:, :, 1] = x"""
    if axis >= array.ndim:
        raise IndexError(
            f"too many indices for array: array is {len(array.shape)}-dimensional, "
            f"but {axis+1} were indexed"
        )
    # TODO: maybe support negative axis with wraparound
    indices = np.index_exp[:] * axis + np.index_exp[index]
    array[indices] = vals


_np_typechar_to_c_typestring = {
    "?": "bool",
    "B": "unsigned char",
    "D": "double complex",
    "F": "float complex",
    "G": "long double complex",
    "H": "unsigned short",
    "I": "unsigned int",
    "L": "unsigned long",
    "Q": "unsigned long long",
    "b": "char",
    "d": "double",
    "e": "half",  # warning: only in CUDA with special support
    "f": "float",
    "g": "long double",
    "h": "short",
    "i": "int",
    "l": "long",
    "q": "long long",
}


def np_dtype_to_c_type(dtype):
    as_char = np.sctype2char(dtype)
    return _np_typechar_to_c_typestring[as_char]


def enum_to_c_template(enum_class):
    res = [f"enum {enum_class.__name__} {{"]
    for field in enum_class:
        res.append(f"\t{field.name} = {field.value},")
    res.append("};")
    return "\n".join(res)


def ceil_div(num, denom):
    return (num + denom - 1) // denom


class RepeatingTimer:
    """A timer which will call callback every interval seconds"""

    def __init__(
        self,
        interval,
        callback,
        start_now=True,
        daemon=True,
    ):
        self.stopped = True
        self.interval = interval
        self.callback = callback
        self.daemonize = daemon
        if start_now:
            self.start()

    def start(self):
        self.stopped = False
        self.wakeup_time = default_timer() + self.interval

        def runner():
            while not self.stopped:
                now = default_timer()
                while now < self.wakeup_time:
                    diff = self.wakeup_time - now
                    time.sleep(diff)
                    if self.stopped:
                        return
                    now = default_timer()
                self.callback()
                self.wakeup_time = default_timer() + self.interval

        self.thread = threading.Thread(target=runner, daemon=self.daemonize)
        self.thread.start()

    def stop(self):
        self.stopped = True


class ExponentialMovingAverage:
    def __init__(self, alpha, use_first_value=True):
        self.alpha = alpha
        self.initialised = not use_first_value
        self.mean = 0

    def update(self, value):
        if self.initialised:
            self.mean += self.alpha * (value - self.mean)
        else:
            self.mean = value
            self.initialised = True

    def get(self):
        return self.mean


class WindowRateTracker:
    def __init__(self, buffer_size=20, time_window=10):
        self.time_window = time_window
        self.buffer_size = buffer_size
        self.deque = collections.deque(maxlen=self.buffer_size)

    def update(self):
        self.deque.append(default_timer())

    def get(self):
        now = default_timer()
        cutoff = now - self.time_window
        try:
            while self.deque[0] < cutoff:
                self.deque.popleft()
        except IndexError:
            return 0
        if len(self.deque) < 2:
            return 0
        if len(self.deque) < self.buffer_size:
            # TODO: estimator avoiding ramp-up of when starting anew
            return len(self.deque) / self.time_window
        else:
            # if going faster than buffer size per time window, look at timestamps
            oldest, newest = self.deque[0], self.deque[-1]
            buffer_span = newest - oldest
            period = buffer_span / (self.buffer_size - 1)
            if (now - newest) < period:
                # no new estimate yet, expecting new event after period
                return 1 / period
            else:
                return self.buffer_size / (now - oldest)


class Stopwatch:
    """Context manager measuring time spent in context.

    Keyword arguments:
    name: if not None, will appear in string representation
          also, if not None, will automatically print self when done
    """

    def __init__(self, name=None):
        self.stop_time = None
        self.name = name

    def __enter__(self):
        self.start_time = default_timer()
        return self

    def __exit__(self, t, v, tb):  # type, value and traceback irrelevant
        self.stop_time = default_timer()
        if self.name is not None:
            print(repr(self))

    @property
    def elapsed(self):
        if self.stop_time is not None:
            return self.stop_time - self.start_time
        else:
            return default_timer() - self.start_time

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        if self.name is None:
            return f"{self.elapsed:.3f} s"
        else:
            return f"{self.name}: {self.elapsed:.3f} s"


class StateContext:
    """What if device state was a stack?"""

    def __init__(self, device, new_state, revert_to=None):
        self.device = device
        if revert_to is None:
            revert_to = device.get("state")
        self.revert_to = revert_to
        self.new_state = new_state

    def __enter__(self):
        if self.new_state is not self.revert_to:
            self.device.updateState(self.new_state)

    def __exit__(self, t, v, tb):
        if self.new_state is not self.revert_to:
            self.device.updateState(self.revert_to)


class NonMonotonicTrainIdWarning(Warning):
    pass


class TrainRatioTracker:
    """Measure how many percent of recent train IDs (from contiguous set) were seen

    The tracker will maintain a queue of buffer_size train IDs going back at most
    buffer_size from latest train ID (depending on calls to get). Call update(train_id)
    when you see a new train and call get to get() the ratio of recent trains seen.

    Updating will raise NonMonotonicTrainIdWarning or LargeTrainIdGapWarning in case
    train ID looks like it's from far in the future or from some time in the past.
    Device using this tracker should decide what to do; maybe call reset.
    """

    def __init__(self, buffer_size=50):
        self._train_id_queue = collections.deque(maxlen=buffer_size)

    def get(self, current_train=None, expected_delay=None):
        """Get the ratio of recent trains based on buffer contents

        If current_train is provided, the train ID span used to compute the ratio is
        from oldest tid in buffer to current_train.  Otherwise, latest update tid is
        used as stand-in for current_train.

        If expected_delay is provided in addition to current_train, will take max of
        latest update tid and current train minus expected delay.
        """
        try:
            if current_train is None:
                current_train = self._train_id_queue[-1]
            elif expected_delay is not None:
                current_train = max(
                    current_train - expected_delay, self._train_id_queue[-1]
                )
            oldest_train = self._train_id_queue[0]
        except IndexError:
            return 0

        return len(self._train_id_queue) * 100 / (current_train - oldest_train + 1)

    def reset(self):
        self._train_id_queue.clear()

    def update(self, train_id):
        # allows same train ID multiple times
        if self._train_id_queue and (last_seen := self._train_id_queue[-1]) > train_id:
            raise NonMonotonicTrainIdWarning(
                "New train ID not greater than last train ID seen! "
                f"New: {train_id}, previous: {last_seen}"
            )
        self._train_id_queue.append(train_id)


class ChainHash:
    """Like read-only ChainMap, but for karabo.bound.Hash(es) instead!"""

    def __init__(self, *hashes):
        self._hashes = hashes

    def __getitem__(self, key):
        for h in self._hashes:
            if h.has(key):
                return h[key]
        raise KeyError()

    def get(self, key):
        return self[key]


class SkippingThrottler:
    def __init__(self, min_period):
        self.min_period = min_period
        self.latest_ts = float("-inf")
        self.lock = threading.Lock()

    def test_and_set(self):
        with self.lock:
            now = default_timer()
            if (now - self.latest_ts) >= self.min_period:
                self.latest_ts = now
                return True
            else:
                return False


class BadPixelValues(enum.IntFlag):
    """The European XFEL Bad Pixel Encoding

    Straight from pycalibration's enum.py"""

    OFFSET_OUT_OF_THRESHOLD = 2 ** 0
    NOISE_OUT_OF_THRESHOLD = 2 ** 1
    OFFSET_NOISE_EVAL_ERROR = 2 ** 2
    NO_DARK_DATA = 2 ** 3
    CI_GAIN_OUT_OF_THRESHOLD = 2 ** 4
    CI_LINEAR_DEVIATION = 2 ** 5
    CI_EVAL_ERROR = 2 ** 6
    FF_GAIN_EVAL_ERROR = 2 ** 7
    FF_GAIN_DEVIATION = 2 ** 8
    FF_NO_ENTRIES = 2 ** 9
    CI2_EVAL_ERROR = 2 ** 10
    VALUE_IS_NAN = 2 ** 11
    VALUE_OUT_OF_RANGE = 2 ** 12
    GAIN_THRESHOLDING_ERROR = 2 ** 13
    DATA_STD_IS_ZERO = 2 ** 14
    ASIC_STD_BELOW_NOISE = 2 ** 15
    INTERPOLATED = 2 ** 16
    NOISY_ADC = 2 ** 17
    OVERSCAN = 2 ** 18
    NON_SENSITIVE = 2 ** 19
    NON_LIN_RESPONSE_REGION = 2 ** 20
    WRONG_GAIN_VALUE = 2 ** 21
    NON_STANDARD_SIZE = 2 ** 22


def downsample_2d(arr, factor, reduction_fun=np.nanmax):
    """Generalization of downsampling from FemDataAssembler

    Expects first two dimensions of arr to be multiple of 2 ** factor
    Useful if you're sitting at home and ssh connection is slow to get full-resolution
    previews."""

    for i in range(factor // 2):
        arr = reduction_fun(
            (
                arr[:-1:2],
                arr[1::2],
            ),
            axis=0,
        )
        arr = reduction_fun(
            (
                arr[:, :-1:2],
                arr[:, 1::2],
            ),
            axis=0,
        )
    return arr


def blocked_view_2d(A, blocks_ss, blocks_fs):
    assert A.shape[0] % blocks_ss == 0
    assert A.shape[1] % blocks_fs == 0
    return A.reshape(
        blocks_ss,
        A.shape[0] // blocks_ss,
        blocks_fs,
        A.shape[1] // blocks_fs,
    ).swapaxes(
        1,
        2,
    )


def quadrant_views(A):
    res = []
    for row in np.vsplit(A, 2):
        res.extend(np.hsplit(row, 2))
    return res


def cell_table_to_string(cell_table):
    return ",".join(map(str, cell_table.ravel())) + ","


def grid_to_cover_shape_with_blocks(full_shape, block_shape):
    return tuple(itertools.starmap(ceil_div, zip(full_shape, block_shape)))


def apply_partial_lut(data, lut, mask, out, missing=np.nan):
    tmp = out.ravel()
    tmp[~mask] = data.ravel()[lut]
    tmp[mask] = missing


def add_unsafe_get(device_class):
    # forward-compatible unsafe_get proposed by @haufs
    if not hasattr(device_class, "unsafe_get"):

        def unsafe_get(self, key):
            """Look up key in device schema quickly, but without consistency locks

            This is only relevant for use in hot path (input handler).  Circumvents the
            locking done by PythonDevice.get. Note that PythonDevice.get does handle
            some special types (by looking at full schema for type information).  In
            particular, device state enum: `self.get("state")` will return a State
            whereas `self.unsafe_get("state")` will return a string. Handle with care!
            """

            # at least until Karabo 2.14, self._parameters is maintained by PythonDevice
            return self._parameters.get(key)

        setattr(device_class, "unsafe_get", unsafe_get)
